//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ CCalculator class header ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// Calculator.h: interface for the CCalculator class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_CALCULATOR_H__BAA54421_D2BD_11D1_BD69_0000C053B4FA__INCLUDED_)
#define AFX_CALCULATOR_H__BAA54421_D2BD_11D1_BD69_0000C053B4FA__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "Emulator.h"
#include "LinkPort.h"	// Added by ClassView

class CCalculator  
{
public:
	CCalculator();
	virtual ~CCalculator();

	// Operations
public:
	CBitmap * GetScreenBitmap();

	virtual bool Reset();
	virtual bool Start();
	virtual bool Stop();
	virtual DWORD Resume();
	virtual DWORD Suspend();

	virtual bool LoadROM( const char * pszFile );
	virtual bool StoreRAM( const char * pszFile, unsigned nVersion = 0x0201 ) = 0;
	virtual bool LoadRAM( const char * pszFile ) = 0;
	
	virtual void Output( unsigned char Port, unsigned char Data ) = 0;
	virtual unsigned char Input( unsigned char Port ) = 0;
	virtual unsigned char LinkInput();
	virtual void LinkOutput(unsigned char Data);
	virtual unsigned short Interrupt();

	reg * GetRegs();
	void SetRegs(reg Regs);

	// Public attributes
public:
	unsigned char m_RomPageNo;
	unsigned m_RomSize;
	unsigned char * m_pMemory;
	unsigned m_cyScreen;
	unsigned m_cxScreen;
	unsigned char m_Model;		// 0x82, 0x83, ...
	reg m_Regs;
	bool volatile m_KeepAlive;
	HWND m_hwndScreen;
	
	//
	// Linking
	//
	unsigned char m_LinkOutput;
	unsigned char volatile m_LinkState;
	bool volatile m_LinkActive;

	// Internal attributes
protected:
	unsigned char m_RomPage;
	unsigned char * m_pScreen;
	unsigned char * m_pRom;
	HANDLE m_hEmulatorThread;
	unsigned char m_VKeys[0x100];

	// Unsorted
public:
	bool m_LinkSending;
	BOOL SetInputThread(DWORD idThread, BOOL fAttach);
	DWORD m_EmulatorThreadID;
	unsigned m_Suspend;
	bool m_RefreshOnInt;
	void StopLink();
	HANDLE m_hInterrupt;
	HANDLE m_hInterruptNoWait;

	CLinkPort m_LinkPort;
	HANDLE m_hLinkThread;
	virtual bool SendFiles( CStringArray * pFiles, HWND hProgressWnd );

};

extern CCalculator * g_pCalculator;

#define COMBINE2(a,b)	(m_VKeys[a] = m_VKeys[b] = (m_VKeys[a] | m_VKeys[b]))
#define COMBINE3(a,b,c)	(m_VKeys[a] = m_VKeys[b] = m_VKeys[c] = (m_VKeys[a] | m_VKeys[b] | m_VKeys[c]))

#ifdef _DEBUG
#define TRACE_PC(pc)	TRACE("%X:%04X  ", g_pCalculator->m_RomPageNo, pc)
#define TRACE_PC_82(pc)	TRACE_PC(pc)
#define TRACE_PC_83(pc)	TRACE_PC(pc)
#endif

#endif // !defined(AFX_CALCULATOR_H__BAA54421_D2BD_11D1_BD69_0000C053B4FA__INCLUDED_)
