//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator   Emulator   (CalcEm) ]
// Source File:  [ Z80 Emulator   implementation ]
// Modified by:  [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.4 ]
//////////////////////////////////////////////////////////////////////////////////
#pragma warning(disable : 4244 4706 4701)

/** Z80: portable Z80 emulator   *******************************/
/**                                                         **/
/**                         Debug.c                         **/
/**                                                         **/
/** This file contains the built-in   debugging routine for     **/
/** the Z80 emulator   which is called on each Z80 step       **/
/** when Trap==TRUE.                                        **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin   1995,1996                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "Z80.h"

// integral size mismatch in argument; conversion supplied
#pragma warning(disable : 4761)

#include <stdio.h>
#include <string.h>

int DAsm(char *S,word A);

char *Mnemonics[256] =
{
  "nop", "ld   bc, #h", "ld   (bc), a", "inc  bc", "inc  b", "dec b", "ld   b, *h", "rlca", 
  "ex   af, af'", "add  hl, bc", "ld   a, (bc)", "dec bc", "inc  c", "dec c", "ld   c, *h", "rrca", 
  "djnz @h", "ld   de, #h", "ld   (de), a", "inc  de", "inc  d", "dec d", "ld   d, *h", "rla", 
  "jr   @h", "add  hl, de", "ld   a, (de)", "dec de", "inc  e", "dec e", "ld   e, *h", "rra", 
  "jr   nz, @h", "ld   hl, #h", "ld   (#h), hl", "inc  hl", "inc  h", "dec h", "ld   h, *h", "daa", 
  "jr   z, @h", "add  hl, hl", "ld   hl, (#h)", "dec hl", "inc  l", "dec l", "ld   l, *h", "cpl", 
  "jr   nc, @h", "ld   sp, #h", "ld   (#h), a", "inc  sp", "inc  (hl)", "dec (hl)", "ld   (hl), *h", "scf", 
  "jr   c, @h", "add  hl, sp", "ld   a, (#h)", "dec sp", "inc  a", "dec a", "ld   a, *h", "ccf", 
  "ld   b, b", "ld   b, c", "ld   b, d", "ld   b, e", "ld   b, h", "ld   b, l", "ld   b, (hl)", "ld   b, a", 
  "ld   c, b", "ld   c, c", "ld   c, d", "ld   c, e", "ld   c, h", "ld   c, l", "ld   c, (hl)", "ld   c, a", 
  "ld   d, b", "ld   d, c", "ld   d, d", "ld   d, e", "ld   d, h", "ld   d, l", "ld   d, (hl)", "ld   d, a", 
  "ld   e, b", "ld   e, c", "ld   e, d", "ld   e, e", "ld   e, h", "ld   e, l", "ld   e, (hl)", "ld   e, a", 
  "ld   h, b", "ld   h, c", "ld   h, d", "ld   h, e", "ld   h, h", "ld   h, l", "ld   h, (hl)", "ld   h, a", 
  "ld   l, b", "ld   l, c", "ld   l, d", "ld   l, e", "ld   l, h", "ld   l, l", "ld   l, (hl)", "ld   l, a", 
  "ld   (hl), b", "ld   (hl), c", "ld   (hl), d", "ld   (hl), e", "ld   (hl), h", "ld   (hl), l", "halt", "ld   (hl), a", 
  "ld   a, b", "ld   a, c", "ld   a, d", "ld   a, e", "ld   a, h", "ld   a, l", "ld   a, (hl)", "ld   a, a", 
  "add  b", "add  c", "add  d", "add  e", "add  h", "add  l", "add  (hl)", "add  a", 
  "adc  b", "adc  c", "adc  d", "adc  e", "adc  h", "adc  l", "adc  (hl)", "adc  a", 
  "sub  b", "sub  c", "sub  d", "sub  e", "sub  h", "sub  l", "sub  (hl)", "sub  a", 
  "sbc  b", "sbc  c", "sbc  d", "sbc  e", "sbc  h", "sbc  l", "sbc  (hl)", "sbc  a", 
  "and  b", "and  c", "and  d", "and  e", "and  h", "and  l", "and  (hl)", "and  a", 
  "xor  b", "xor  c", "xor  d", "xor  e", "xor  h", "xor  l", "xor  (hl)", "xor  a", 
  "or   b", "or   c", "or   d", "or   e", "or   h", "or   l", "or   (hl)", "or   a", 
  "cp   b", "cp   c", "cp   d", "cp   e", "cp   h", "cp   l", "cp   (hl)", "cp   a", 
  "ret  nz", "pop  bc", "jp   nz, #h", "jp   #h", "call nz, #h", "push bc", "add  *h", "rst  00h", 
  "ret  z", "ret", "jp   z, #h", "pfx_cb", "call z, #h", "call #h", "adc  *h", "rst  08h", 
  "ret  nc", "pop  de", "jp   nc, #h", "out  a, (*h)", "call nc, #h", "push de", "sub  *h", "rst  10h", 
  "ret  c", "exx", "jp   c, #h", "in   a, (*h)", "call c, #h", "pfx_dd", "sbc  *h", "rst  18h", 
  "ret  po", "pop  hl", "jp   po, #h", "ex   hl, (sp)", "call po, #h", "push hl", "and  *h", "rst  20h", 
  "ret  pe", "ld   pc, hl", "jp   pe, #h", "ex   de, hl", "call pe, #h", "pfx_ed", "xor  *h", "rst  28h", 
  "ret  p", "pop  af", "jp   p, #h", "di", "call p, #h", "push af", "or   *h", "rst  30h", 
  "ret  m", "ld   sp, hl", "jp   m, #h", "ei", "call m, #h", "pfx_fd", "cp   *h", "rst  38h"
};

char *MnemonicsCB[256] =
{
  "rlc  b", "rlc  c", "rlc  d", "rlc  e", "rlc  h", "rlc  l", "rlc  (hl)", "rlc  a", 
  "rrc  b", "rrc  c", "rrc  d", "rrc  e", "rrc  h", "rrc  l", "rrc  (hl)", "rrc  a", 
  "rl   b", "rl   c", "rl   d", "rl   e", "rl   h", "rl   l", "rl   (hl)", "rl   a", 
  "rr   b", "rr   c", "rr   d", "rr   e", "rr   h", "rr   l", "rr   (hl)", "rr   a", 
  "sla  b", "sla  c", "sla  d", "sla  e", "sla  h", "sla  l", "sla  (hl)", "sla  a", 
  "sra  b", "sra  c", "sra  d", "sra  e", "sra  h", "sra  l", "sra  (hl)", "sra  a", 
  "sll  b", "sll  c", "sll  d", "sll  e", "sll  h", "sll  l", "sll  (hl)", "sll  a", 
  "srl  b", "srl  c", "srl  d", "srl  e", "srl  h", "srl  l", "srl  (hl)", "srl  a", 
  "bit  0, b", "bit  0, c", "bit  0, d", "bit  0, e", "bit  0, h", "bit  0, l", "bit  0, (hl)", "bit  0, a", 
  "bit  1, b", "bit  1, c", "bit  1, d", "bit  1, e", "bit  1, h", "bit  1, l", "bit  1, (hl)", "bit  1, a", 
  "bit  2, b", "bit  2, c", "bit  2, d", "bit  2, e", "bit  2, h", "bit  2, l", "bit  2, (hl)", "bit  2, a", 
  "bit  3, b", "bit  3, c", "bit  3, d", "bit  3, e", "bit  3, h", "bit  3, l", "bit  3, (hl)", "bit  3, a", 
  "bit  4, b", "bit  4, c", "bit  4, d", "bit  4, e", "bit  4, h", "bit  4, l", "bit  4, (hl)", "bit  4, a", 
  "bit  5, b", "bit  5, c", "bit  5, d", "bit  5, e", "bit  5, h", "bit  5, l", "bit  5, (hl)", "bit  5, a", 
  "bit  6, b", "bit  6, c", "bit  6, d", "bit  6, e", "bit  6, h", "bit  6, l", "bit  6, (hl)", "bit  6, a", 
  "bit  7, b", "bit  7, c", "bit  7, d", "bit  7, e", "bit  7, h", "bit  7, l", "bit  7, (hl)", "bit  7, a", 
  "res  0, b", "res  0, c", "res  0, d", "res  0, e", "res  0, h", "res  0, l", "res  0, (hl)", "res  0, a", 
  "res  1, b", "res  1, c", "res  1, d", "res  1, e", "res  1, h", "res  1, l", "res  1, (hl)", "res  1, a", 
  "res  2, b", "res  2, c", "res  2, d", "res  2, e", "res  2, h", "res  2, l", "res  2, (hl)", "res  2, a", 
  "res  3, b", "res  3, c", "res  3, d", "res  3, e", "res  3, h", "res  3, l", "res  3, (hl)", "res  3, a", 
  "res  4, b", "res  4, c", "res  4, d", "res  4, e", "res  4, h", "res  4, l", "res  4, (hl)", "res  4, a", 
  "res  5, b", "res  5, c", "res  5, d", "res  5, e", "res  5, h", "res  5, l", "res  5, (hl)", "res  5, a", 
  "res  6, b", "res  6, c", "res  6, d", "res  6, e", "res  6, h", "res  6, l", "res  6, (hl)", "res  6, a", 
  "res  7, b", "res  7, c", "res  7, d", "res  7, e", "res  7, h", "res  7, l", "res  7, (hl)", "res  7, a", 
  "set  0, b", "set  0, c", "set  0, d", "set  0, e", "set  0, h", "set  0, l", "set  0, (hl)", "set  0, a", 
  "set  1, b", "set  1, c", "set  1, d", "set  1, e", "set  1, h", "set  1, l", "set  1, (hl)", "set  1, a", 
  "set  2, b", "set  2, c", "set  2, d", "set  2, e", "set  2, h", "set  2, l", "set  2, (hl)", "set  2, a", 
  "set  3, b", "set  3, c", "set  3, d", "set  3, e", "set  3, h", "set  3, l", "set  3, (hl)", "set  3, a", 
  "set  4, b", "set  4, c", "set  4, d", "set  4, e", "set  4, h", "set  4, l", "set  4, (hl)", "set  4, a", 
  "set  5, b", "set  5, c", "set  5, d", "set  5, e", "set  5, h", "set  5, l", "set  5, (hl)", "set  5, a", 
  "set  6, b", "set  6, c", "set  6, d", "set  6, e", "set  6, h", "set  6, l", "set  6, (hl)", "set  6, a", 
  "set  7, b", "set  7, c", "set  7, d", "set  7, e", "set  7, h", "set  7, l", "set  7, (hl)", "set  7, a"
};

char *MnemonicsED[256] =
{
  "db   edh, 00h", "db   edh, 01h", "db   edh, 02h", "db   edh, 03h", 
  "db   edh, 04h", "db   edh, 05h", "db   edh, 06h", "db   edh, 07h", 
  "db   edh, 08h", "db   edh, 09h", "db   edh, 0ah", "db   edh, 0bh", 
  "db   edh, 0ch", "db   edh, 0dh", "db   edh, 0eh", "db   edh, 0fh", 
  "db   edh, 10h", "db   edh, 11h", "db   edh, 12h", "db   edh, 13h", 
  "db   edh, 14h", "db   edh, 15h", "db   edh, 16h", "db   edh, 17h", 
  "db   edh, 18h", "db   edh, 19h", "db   edh, 1ah", "db   edh, 1bh", 
  "db   edh, 1ch", "db   edh, 1dh", "db   edh, 1eh", "db   edh, 1fh", 
  "db   edh, 20h", "db   edh, 21h", "db   edh, 22h", "db   edh, 23h", 
  "db   edh, 24h", "db   edh, 25h", "db   edh, 26h", "db   edh, 27h", 
  "db   edh, 28h", "db   edh, 29h", "db   edh, 2ah", "db   edh, 2bh", 
  "db   edh, 2ch", "db   edh, 2dh", "db   edh, 2eh", "db   edh, 2fh", 
  "db   edh, 30h", "db   edh, 31h", "db   edh, 32h", "db   edh, 33h", 
  "db   edh, 34h", "db   edh, 35h", "db   edh, 36h", "db   edh, 37h", 
  "db   edh, 38h", "db   edh, 39h", "db   edh, 3ah", "db   edh, 3bh", 
  "db   edh, 3ch", "db   edh, 3dh", "db   edh, 3eh", "db   edh, 3fh", 
  "in   b, (c)", "out  (c), b", "sbc  hl, bc", "ld   (#h), bc", 
  "db   edh, 44h", "retn", "im   0", "ld   i, a", 
  "in   c, (c)", "out  (c), c", "adc  hl, bc", "ld   bc, (#h)", 
  "db   edh, 4ch", "reti", "db   edh, 4eh", "ld   r, a", 
  "in   d, (c)", "out  (c), d", "sbc  hl, de", "ld   (#h), de", 
  "db   edh, 54h", "db   edh, 55h", "im   1", "ld   a, i", 
  "in   e, (c)", "out  (c), e", "adc  hl, de", "ld   de, (#h)", 
  "db   edh, 5ch", "db   edh, 5dh", "im   2", "ld   a, r", 
  "in   h, (c)", "out  (c), h", "sbc  hl, hl", "ld   (#h), hl", 
  "db   edh, 64h", "db   edh, 65h", "db   edh, 66h", "rrd", 
  "in   l, (c)", "out  (c), l", "adc  hl, hl", "ld   hl, (#h)", 
  "db   edh, 6ch", "db   edh, 6dh", "db   edh, 6eh", "rld", 
  "in   f, (c)", "db   edh, 71h", "sbc  hl, sp", "ld   (#h), sp", 
  "db   edh, 74h", "db   edh, 75h", "db   edh, 76h", "db   edh, 77h", 
  "in   a, (c)", "out  (c), a", "adc  hl, sp", "ld   sp, (#h)", 
  "db   edh, 7ch", "db   edh, 7dh", "db   edh, 7eh", "db   edh, 7fh", 
  "db   edh, 80h", "db   edh, 81h", "db   edh, 82h", "db   edh, 83h", 
  "db   edh, 84h", "db   edh, 85h", "db   edh, 86h", "db   edh, 87h", 
  "db   edh, 88h", "db   edh, 89h", "db   edh, 8ah", "db   edh, 8bh", 
  "db   edh, 8ch", "db   edh, 8dh", "db   edh, 8eh", "db   edh, 8fh", 
  "db   edh, 90h", "db   edh, 91h", "db   edh, 92h", "db   edh, 93h", 
  "db   edh, 94h", "db   edh, 95h", "db   edh, 96h", "db   edh, 97h", 
  "db   edh, 98h", "db   edh, 99h", "db   edh, 9ah", "db   edh, 9bh", 
  "db   edh, 9ch", "db   edh, 9dh", "db   edh, 9eh", "db   edh, 9fh", 
  "ldi", "cpi", "ini", "outi", 
  "db   edh, a4h", "db   edh, a5h", "db   edh, a6h", "db   edh, a7h", 
  "ldd", "cpd", "ind", "outd", 
  "db   edh, ach", "db   edh, adh", "db   edh, aeh", "db   edh, afh", 
  "ldir", "cpir", "inir", "otir", 
  "db   edh, b4h", "db   edh, b5h", "db   edh, b6h", "db   edh, b7h", 
  "lddr", "cpdr", "indr", "otdr", 
  "db   edh, bch", "db   edh, bdh", "db   edh, beh", "db   edh, bfh", 
  "db   edh, c0h", "db   edh, c1h", "db   edh, c2h", "db   edh, c3h", 
  "db   edh, c4h", "db   edh, c5h", "db   edh, c6h", "db   edh, c7h", 
  "db   edh, c8h", "db   edh, c9h", "db   edh, cah", "db   edh, cbh", 
  "db   edh, cch", "db   edh, cdh", "db   edh, ceh", "db   edh, cfh", 
  "db   edh, d0h", "db   edh, d1h", "db   edh, d2h", "db   edh, d3h", 
  "db   edh, d4h", "db   edh, d5h", "db   edh, d6h", "db   edh, d7h", 
  "db   edh, d8h", "db   edh, d9h", "db   edh, dah", "db   edh, dbh", 
  "db   edh, dch", "db   edh, ddh", "db   edh, deh", "db   edh, dfh", 
  "db   edh, e0h", "db   edh, e1h", "db   edh, e2h", "db   edh, e3h", 
  "db   edh, e4h", "db   edh, e5h", "db   edh, e6h", "db   edh, e7h", 
  "db   edh, e8h", "db   edh, e9h", "db   edh, eah", "db   edh, ebh", 
  "db   edh, ech", "db   edh, edh", "db   edh, eeh", "db   edh, efh", 
  "db   edh, f0h", "db   edh, f1h", "db   edh, f2h", "db   edh, f3h", 
  "db   edh, f4h", "db   edh, f5h", "db   edh, f6h", "db   edh, f7h", 
  "db   edh, f8h", "db   edh, f9h", "db   edh, fah", "db   edh, fbh", 
  "db   edh, fch", "db   edh, fdh", "db   edh, feh", "db   edh, ffh"
};

char *MnemonicsXX[256] =
{
  "nop", "ld   bc, #h", "ld   (bc), a", "inc  bc", "inc  b", "dec b", "ld   b, *h", "rlca", 
  "ex     af, af'", "add  i%, bc", "ld   a, (bc)", "dec bc", "inc  c", "dec c", "ld   c, *h", "rrca", 
  "djnz @h", "ld   de, #h", "ld   (de), a", "inc  de", "inc  d", "dec d", "ld   d, *h", "rla", 
  "jr    @h", "add   i%, de", "ld   a, (de)", "dec de", "inc  e", "dec e", "ld   e, *h", "rra", 
  "jr    nz, @h", "ld   i%, #h", "ld   (#h), i%", "inc  i%", "inc  i%h", "dec i%h", "ld   i%h, *h", "daa", 
  "jr    z, @h", "add  i%, i%", "ld   i%, (#h)", "dec i%", "inc  i%l", "dec i%l", "ld   i%l, *h", "cpl", 
  "jr    nc, @h", "ld   sp, #h", "ld   (#h), a", "inc  sp", "inc  (i%+^h)", "dec (i%+^h)", "ld   (i%+^h), *h", "scf", 
  "jr    c, @h", "add  i%, sp", "ld   a, (#h)", "dec sp", "inc  a", "dec a", "ld   a, *h", "ccf", 
  "ld   b, b", "ld   b, c", "ld   b, d", "ld   b, e", "ld   b, i%h", "ld   b, i%l", "ld   b, (i%+^h)", "ld   b, a", 
  "ld   c, b", "ld   c, c", "ld   c, d", "ld   c, e", "ld   c, i%h", "ld   c, i%l", "ld   c, (i%+^h)", "ld   c, a", 
  "ld   d, b", "ld   d, c", "ld   d, d", "ld   d, e", "ld   d, i%h", "ld   d, i%l", "ld   d, (i%+^h)", "ld   d, a", 
  "ld   e, b", "ld   e, c", "ld   e, d", "ld   e, e", "ld   e, i%h", "ld   e, i%l", "ld   e, (i%+^h)", "ld   e, a", 
  "ld   i%h, b", "ld   i%h, c", "ld   i%h, d", "ld   i%h, e", "ld   i%h, i%h", "ld   i%h, i%l", "ld   h, (i%+^h)", "ld   i%h, a", 
  "ld   i%l, b", "ld   i%l, c", "ld   i%l, d", "ld   i%l, e", "ld   i%l, i%h", "ld   i%l, i%l", "ld   l, (i%+^h)", "ld   i%l, a", 
  "ld   (i%+^h), b", "ld   (i%+^h), c", "ld   (i%+^h), d", "ld   (i%+^h), e", "ld   (i%+^h), h", "ld   (i%+^h), l", "halt", "ld   (i%+^h), a", 
  "ld   a, b", "ld   a, c", "ld   a, d", "ld   a, e", "ld   a, i%h", "ld   a, i%l", "ld   a, (i%+^h)", "ld   a, a", 
  "add  b", "add  c", "add  d", "add  e", "add  i%h", "add  i%l", "add  (i%+^h)", "add  a", 
  "adc  b", "adc  c", "adc  d", "adc  e", "adc  i%h", "adc  i%l", "adc  (i%+^h)", "adc, a", 
  "sub  b", "sub  c", "sub  d", "sub  e", "sub  i%h", "sub  i%l", "sub  (i%+^h)", "sub  a", 
  "sbc  b", "sbc  c", "sbc  d", "sbc  e", "sbc  i%h", "sbc  i%l", "sbc  (i%+^h)", "sbc  a", 
  "and  b", "and  c", "and  d", "and  e", "and  i%h", "and  i%l", "and  (i%+^h)", "and  a", 
  "xor  b", "xor  c", "xor  d", "xor  e", "xor  i%h", "xor  i%l", "xor  (i%+^h)", "xor  a", 
  "or   b", "or   c", "or   d", "or   e", "or   i%h", "or   i%l", "or   (i%+^h)", "or   a", 
  "cp   b", "cp   c", "cp   d", "cp   e", "cp   i%h", "cp   i%l", "cp   (i%+^h)", "cp   a", 
  "ret  nz", "pop  bc", "jp   nz, #h", "jp   #h", "call nz, #h", "push bc", "add  *h", "rst  00h", 
  "ret  z", "ret", "jp   z, #h", "pfx_cb", "call z, #h", "call #h", "adc  *h", "rst  08h", 
  "ret  nc", "pop  de", "jp   nc, #h", "out  a, (*h)", "call nc, #h", "push de", "sub  *h", "rst  10h", 
  "ret  c", "exx", "jp   c, #h", "in   a, (*h)", "call c, #h", "pfx_dd", "sbc  *h", "rst  18h", 
  "ret  po", "pop  i%", "jp   po, #h", "ex   i%, (sp)", "call po, #h", "push i%", "and  *h", "rst  20h", 
  "ret  pe", "ld   pc, i%", "jp   pe, #h", "ex   de, i%", "call pe, #h", "pfx_ed", "xor  *h", "rst  28h", 
  "ret  p", "pop  af", "jp   p, #h", "di", "call p, #h", "push af", "or   *h", "rst  30h", 
  "ret  m", "ld   sp, i%", "jp   m, #h", "ei", "call m, #h", "pfx_fd", "cp   *h", "rst  38h"
};

char *MnemonicsXCB[256] =
{
  "rlc  b", "rlc  c", "rlc  d", "rlc  e", "rlc  h", "rlc  l", "rlc  (i%@h)", "rlc  a", 
  "rrc  b", "rrc  c", "rrc  d", "rrc  e", "rrc  h", "rrc  l", "rrc  (i%@h)", "rrc  a", 
  "rl   b", "rl   c", "rl   d", "rl   e", "rl   h", "rl   l", "rl   (i%@h)", "rl   a", 
  "rr   b", "rr   c", "rr   d", "rr   e", "rr   h", "rr   l", "rr   (i%@h)", "rr   a", 
  "sla  b", "sla  c", "sla  d", "sla  e", "sla  h", "sla  l", "sla  (i%@h)", "sla  a", 
  "sra  b", "sra  c", "sra  d", "sra  e", "sra  h", "sra  l", "sra  (i%@h)", "sra  a", 
  "sll  b", "sll  c", "sll  d", "sll  e", "sll  h", "sll  l", "sll  (i%@h)", "sll  a", 
  "srl  b", "srl  c", "srl  d", "srl  e", "srl  h", "srl  l", "srl  (i%@h)", "srl  a", 
  "bit  0, b", "bit  0, c", "bit  0, d", "bit  0, e", "bit  0, h", "bit  0, l", "bit  0, (i%@h)", "bit  0, a", 
  "bit  1, b", "bit  1, c", "bit  1, d", "bit  1, e", "bit  1, h", "bit  1, l", "bit  1, (i%@h)", "bit  1, a", 
  "bit  2, b", "bit  2, c", "bit  2, d", "bit  2, e", "bit  2, h", "bit  2, l", "bit  2, (i%@h)", "bit  2, a", 
  "bit  3, b", "bit  3, c", "bit  3, d", "bit  3, e", "bit  3, h", "bit  3, l", "bit  3, (i%@h)", "bit  3, a", 
  "bit  4, b", "bit  4, c", "bit  4, d", "bit  4, e", "bit  4, h", "bit  4, l", "bit  4, (i%@h)", "bit  4, a", 
  "bit  5, b", "bit  5, c", "bit  5, d", "bit  5, e", "bit  5, h", "bit  5, l", "bit  5, (i%@h)", "bit  5, a", 
  "bit  6, b", "bit  6, c", "bit  6, d", "bit  6, e", "bit  6, h", "bit  6, l", "bit  6, (i%@h)", "bit  6, a", 
  "bit  7, b", "bit  7, c", "bit  7, d", "bit  7, e", "bit  7, h", "bit  7, l", "bit  7, (i%@h)", "bit  7, a", 
  "res  0, b", "res  0, c", "res  0, d", "res  0, e", "res  0, h", "res  0, l", "res  0, (i%@h)", "res  0, a", 
  "res  1, b", "res  1, c", "res  1, d", "res  1, e", "res  1, h", "res  1, l", "res  1, (i%@h)", "res  1, a", 
  "res  2, b", "res  2, c", "res  2, d", "res  2, e", "res  2, h", "res  2, l", "res  2, (i%@h)", "res  2, a", 
  "res  3, b", "res  3, c", "res  3, d", "res  3, e", "res  3, h", "res  3, l", "res  3, (i%@h)", "res  3, a", 
  "res  4, b", "res  4, c", "res  4, d", "res  4, e", "res  4, h", "res  4, l", "res  4, (i%@h)", "res  4, a", 
  "res  5, b", "res  5, c", "res  5, d", "res  5, e", "res  5, h", "res  5, l", "res  5, (i%@h)", "res  5, a", 
  "res  6, b", "res  6, c", "res  6, d", "res  6, e", "res  6, h", "res  6, l", "res  6, (i%@h)", "res  6, a", 
  "res  7, b", "res  7, c", "res  7, d", "res  7, e", "res  7, h", "res  7, l", "res  7, (i%@h)", "res  7, a", 
  "set  0, b", "set  0, c", "set  0, d", "set  0, e", "set  0, h", "set  0, l", "set  0, (i%@h)", "set  0, a", 
  "set  1, b", "set  1, c", "set  1, d", "set  1, e", "set  1, h", "set  1, l", "set  1, (i%@h)", "set  1, a", 
  "set  2, b", "set  2, c", "set  2, d", "set  2, e", "set  2, h", "set  2, l", "set  2, (i%@h)", "set  2, a", 
  "set  3, b", "set  3, c", "set  3, d", "set  3, e", "set  3, h", "set  3, l", "set  3, (i%@h)", "set  3, a", 
  "set  4, b", "set  4, c", "set  4, d", "set  4, e", "set  4, h", "set  4, l", "set  4, (i%@h)", "set  4, a", 
  "set  5, b", "set  5, c", "set  5, d", "set  5, e", "set  5, h", "set  5, l", "set  5, (i%@h)", "set  5, a", 
  "set  6, b", "set  6, c", "set  6, d", "set  6, e", "set  6, h", "set  6, l", "set  6, (i%@h)", "set  6, a", 
  "set  7, b", "set  7, c", "set  7, d", "set  7, e", "set  7, h", "set  7, l", "set  7, (i%@h)", "set  7, a"
};

int DAsm(char *S,word A)
{
  char R[128],H[10],C,*T,*P;
  byte J,Offset;
  word B;

  B=A;C='\0';J=0;

  switch(ReadMemory(B))
  {
    case 0xCB: B++;T=MnemonicsCB[ReadMemory(B++)];break;
    case 0xED: B++;T=MnemonicsED[ReadMemory(B++)];break;
    case 0xDD: B++;C='x';
               if(ReadMemory(B)!=0xCB) T=MnemonicsXX[ReadMemory(B++)];
               else
               { B++;Offset=ReadMemory(B++);J=1;T=MnemonicsXCB[ReadMemory(B++)]; }
               break;
    case 0xFD: B++;C='y';
               if(ReadMemory(B)!=0xCB) T=MnemonicsXX[ReadMemory(B++)];
               else
               { B++;Offset=ReadMemory(B++);J=1;T=MnemonicsXCB[ReadMemory(B++)]; }
               break;
    default:   T=Mnemonics[ReadMemory(B++)];
  }

  if(P=strchr(T,'^'))
  {
    strncpy(R,T,P-T);R[P-T]='\0';
    sprintf(H,"%02x",ReadMemory(B++));
    strcat(R,H);strcat(R,P+1);
  }
  else strcpy(R,T);
  if(P=strchr(R,'%')) *P=C;

  if(P=strchr(R,'*'))
  {
    strncpy(S,R,P-R);S[P-R]='\0';
    sprintf(H,"%02x",ReadMemory(B++));
    strcat(S,H);strcat(S,P+1);
  }
  else
    if(P=strchr(R,'@'))
    {
      strncpy(S,R,P-R);S[P-R]='\0';
      if(!J) Offset=ReadMemory(B++);
      strcat(S,Offset&0x80? "-":"+");
      J=Offset&0x80? 256-Offset:Offset;
      sprintf(H,"%02x",J);
      strcat(S,H);strcat(S,P+1);
    }
    else
      if(P=strchr(R,'#'))
      {
        strncpy(S,R,P-R);S[P-R]='\0';
        sprintf(H,"%04x",ReadMemory(B)+256*ReadMemory(B+1));
        strcat(S,H);strcat(S,P+1);
        B+=2;
      }
      else strcpy(S,R);

  return(B-A);
}

#ifdef DEBUG

void __stdcall
OutputDebugStringA(
    const char * lpOutputString
    );


/*** Single-step debugger ****************************/
/*** This function should   exist if DEBUG is        ***/
/*** #defined. When Trace=TRUE, it is called after ***/
/*** each command  executed by the CPU and  given    ***/
/*** a pointer to the register file.               ***/
/*****************************************************/
void Debug(reg *R)
{
	char temp[1024];

	static char Flags[8] = "SZ.H.PNC";
  char S[128],T[10];
  byte J,I;

 static word pc = 0xffff;

	if( pc != R->PC.W )
	{
		pc = R->PC.W;

  DAsm(S,R->PC.W);

  for(J=0,I=R->AF.B.l;J<8;J++,I<<=1) T[J]=I&0x80? Flags[J]:'.';
  T[8]='\0';

  sprintf
  (
	temp,
    "AF:%04X HL:%04X DE:%04X BC:%04X PC:%04X SP:%04X IX:%04X IY:%04X\n",
    R->AF.W,R->HL.W,R->DE.W,R->BC.W,R->PC.W,R->SP.W,R->IX.W,R->IY.W
  ); 

	OutputDebugStringA(temp);

  sprintf
  ( 
	temp,
    "AT PC: [%02X - %s]   AT SP: [%02X]   FLAGS: [%s]\n\n",
    ReadMemory(R->PC.W),S,ReadMemory(R->SP.W)+ReadMemory(R->SP.W+1)*256,T
  );

	OutputDebugStringA(temp);

	}

//  fflush(stdout);
}

#endif
