//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Z80 Emulator implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "Calculator.h"

UINT wm_Refresh = ::RegisterWindowMessage("CalcEm_Refresh");

//============================================================================
//
// EmulatorThread
//
//============================================================================
UINT __cdecl EmulatorThread(PVOID pParam)
{
	ASSERT(pParam);
	CCalculator * pCalculator = (CCalculator *)pParam;

#if 1
	RunZ80( (unsigned char *)&pCalculator->m_KeepAlive );
#else
	while( pCalculator->m_KeepAlive )
		;
#endif

	return 0;
}

//============================================================================
//
// Read byte from I/O port
//
//============================================================================
byte Input(byte Port)
{
	return g_pCalculator->Input(Port);
}

//============================================================================
//
// Write byte to I/O port
//
//============================================================================
void Output(byte Port,byte Value)
{
	g_pCalculator->Output(Port, Value);	
}

//============================================================================
//
// Interrupt handler
//
//============================================================================
word Interrupt(void)
{
	return g_pCalculator->Interrupt();
}

//============================================================================
//
// Write byte to memory
//
//============================================================================
void WriteMemory(word A,byte V)
{
	if( A & 0x8000 )
		g_pCalculator->m_pMemory[A] = V;
#ifdef _DEBUG
	else
	{
		reg Regs;
		Z80_GetRegs(&Regs);
		TRACE_PC(Regs.PC.W);
		TRACE("Memory write attempted at %04X.\n", A);

		//
		// "Normal" ROM writes:
		//	0x0000
		//			TI-82 ROM 18.0 (and probably all other versions too)
		//			TI-83 ROM 1.0600 (and probably all other versions too)
		//			This occurs when the ROM is clearing RAM during memory reset
		//
		//	0x7ff8 to 0x7fff
		//			This occurs each time Crash 1.3 is loading!
		//			(But I'm not sure if it's Crash's own code...)
		//
		// All other ROM memory write attempts should be watched carefully!!!
		//

		// Write a stack dump and stuff...
		if( A != 0 && A < 0x7ff8 )
		{
			// Save/restore suspend count to allow ASSERT dlg show!!!
			unsigned temp = g_pCalculator->m_Suspend;

			g_pCalculator->m_Suspend = (unsigned)-1;

			ASSERT(!A);

			CFile file("stack.bin", CFile::modeCreate | CFile::modeWrite);
			file.Write( g_pCalculator->m_pMemory + Regs.SP.W, 0x10000 - Regs.SP.W );

			g_pCalculator->m_Suspend = temp;
		}
	}
#endif
}

//============================================================================
//
// Read byte to memory
//
//============================================================================
byte ReadMemory(word A)
{
	return g_pCalculator->m_pMemory[A];
}



static void errhandler( PSTR pszError, HWND hWnd )
{
	DWORD dwError = GetLastError();
	TRACE("Bitmap errhandler: %s\n", pszError);
	throw HRESULT_FROM_WIN32( dwError );
}

static PBITMAPINFO CreateBitmapInfoStruct(HWND hwnd, HBITMAP hBmp) { 
    BITMAP bmp; 
    PBITMAPINFO pbmi; 
    WORD    cClrBits; 
 
    /* Retrieve the bitmap's color format, width, and height. */ 
 
    if (!GetObject(hBmp, sizeof(BITMAP), (LPSTR)&bmp)) 
        errhandler("GetObject", hwnd); 
 
 
    /* Convert the color format to a count of bits. */ 
 
    cClrBits = (WORD)(bmp.bmPlanes * bmp.bmBitsPixel); 
 
    if (cClrBits == 1) 
        cClrBits = 1; 
    else if (cClrBits <= 4) 
        cClrBits = 4; 
    else if (cClrBits <= 8) 
        cClrBits = 8; 
    else if (cClrBits <= 16) 
        cClrBits = 16; 
    else if (cClrBits <= 24) 
        cClrBits = 24; 
    else 
        cClrBits = 32; 
 
    /* 
     * Allocate memory for the BITMAPINFO structure. (This structure 
     * contains a BITMAPINFOHEADER structure and an array of RGBQUAD data 
     * structures.) 
     */ 
 
    if (cClrBits != 24) 
         pbmi = (PBITMAPINFO) LocalAlloc(LPTR, 
                    sizeof(BITMAPINFOHEADER) + 
                    sizeof(RGBQUAD) * (2^cClrBits)); 
 
    /* 
     * There is no RGBQUAD array for the 24-bit-per-pixel format. 
     */ 
 
    else 
         pbmi = (PBITMAPINFO) LocalAlloc(LPTR, 
                    sizeof(BITMAPINFOHEADER)); 
 
 
 
    /* Initialize the fields in the BITMAPINFO structure. */ 
 
    pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
    pbmi->bmiHeader.biWidth = bmp.bmWidth; 
    pbmi->bmiHeader.biHeight = bmp.bmHeight; 
    pbmi->bmiHeader.biPlanes = bmp.bmPlanes; 
    pbmi->bmiHeader.biBitCount = bmp.bmBitsPixel; 
    if (cClrBits < 24) 
        pbmi->bmiHeader.biClrUsed = 2^cClrBits; 
 
 
    /* If the bitmap is not compressed, set the BI_RGB flag. */ 
 
    pbmi->bmiHeader.biCompression = BI_RGB; 
 
    /* 
     * Compute the number of bytes in the array of color 
     * indices and store the result in biSizeImage. 
     */ 
 
    pbmi->bmiHeader.biSizeImage = (pbmi->bmiHeader.biWidth + 7) /8 
                                  * pbmi->bmiHeader.biHeight 
                                  * cClrBits; 
 
    /* 
     * Set biClrImportant to 0, indicating that all of the 
     * device colors are important. 
     */ 
 
    pbmi->bmiHeader.biClrImportant = 0; 
 
    return pbmi; 
 
} 

#define MAXWRITE ((DWORD)0xffffffff)

void CreateBMPFile(HWND hwnd, LPCTSTR pszFile, PBITMAPINFO pbi, 
                  HBITMAP hBMP, HDC hDC) 
 { 
 
    HANDLE hf;                  /* file handle */ 
    BITMAPFILEHEADER hdr;       /* bitmap file-header */ 
    PBITMAPINFOHEADER pbih;     /* bitmap info-header */ 
    LPBYTE lpBits;              /* memory pointer */ 
    DWORD dwTotal;              /* total count of bytes */ 
    DWORD cb;                   /* incremental count of bytes */ 
    BYTE *hp;                   /* byte pointer */ 
    DWORD dwTmp; 
 
 
    pbih = (PBITMAPINFOHEADER) pbi; 
    lpBits = (LPBYTE) GlobalAlloc(GMEM_FIXED, pbih->biSizeImage);
    if (!lpBits) 
         errhandler("GlobalAlloc", hwnd); 
 
    /* 
     * Retrieve the color table (RGBQUAD array) and the bits 
     * (array of palette indices) from the DIB. 
     */ 
 
    if (!GetDIBits(hDC, hBMP, 0, (WORD) pbih->biHeight, 
                   lpBits, pbi, DIB_RGB_COLORS)) 
        errhandler("GetDIBits", hwnd); 
 
    /* Create the .BMP file. */ 
 
    hf = CreateFile(pszFile, 
                   GENERIC_READ | GENERIC_WRITE, 
                   (DWORD) 0, 
                   (LPSECURITY_ATTRIBUTES) NULL, 
                   CREATE_ALWAYS, 
                   FILE_ATTRIBUTE_NORMAL, 
                   (HANDLE) NULL); 
 
    if (hf == INVALID_HANDLE_VALUE) 
        errhandler("CreateFile", hwnd); 
 
    hdr.bfType = 0x4d42;        /* 0x42 = "B" 0x4d = "M" */ 
 
    /* Compute the size of the entire file. */ 
 
    hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) + 
                 pbih->biSize + pbih->biClrUsed 
                 * sizeof(RGBQUAD) + pbih->biSizeImage); 
 
    hdr.bfReserved1 = 0; 
    hdr.bfReserved2 = 0; 
 
    /* Compute the offset to the array of color indices. */ 
 
    hdr.bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER) + 
                    pbih->biSize + pbih->biClrUsed 
                    * sizeof (RGBQUAD); 
 
    /* Copy the BITMAPFILEHEADER into the .BMP file. */ 
 
    if (!WriteFile(hf, (LPVOID) &hdr, sizeof(BITMAPFILEHEADER), 
       (LPDWORD) &dwTmp, (LPOVERLAPPED) NULL)) 
       errhandler("WriteFile", hwnd); 
 
	//
	// "Invert" palette
	//
	RGBQUAD temp;
	temp = pbi->bmiColors[0];
	pbi->bmiColors[0] = pbi->bmiColors[1];
	pbi->bmiColors[1] = temp;

    /* Copy the BITMAPINFOHEADER and RGBQUAD array into the file. */ 
 
    if (!WriteFile(hf, (LPVOID) pbih, sizeof(BITMAPINFOHEADER) 
                  + pbih->biClrUsed * sizeof (RGBQUAD), 
                  (LPDWORD) &dwTmp, (LPOVERLAPPED) NULL)) 
       errhandler("WriteFile", hwnd); 
 
    /* Copy the array of color indices into the .BMP file. */ 
 
    dwTotal = cb = pbih->biSizeImage; 
    hp = lpBits;
//    while (cb > MAXWRITE)  { 
//            if (!WriteFile(hf, (LPSTR) hp, (int) MAXWRITE, 
//                          (LPDWORD) &dwTmp, (LPOVERLAPPED) NULL)) 
//                errhandler("WriteFile", hwnd); 
//            cb-= MAXWRITE; 
//            hp += MAXWRITE; 
//    } 
    if (!WriteFile(hf, (LPSTR) hp, (int) cb, 
         (LPDWORD) &dwTmp, (LPOVERLAPPED) NULL)) 
           errhandler("WriteFile", hwnd); 
 
    /* Close the .BMP file. */ 
 
    if (!CloseHandle(hf)) 
           errhandler("CloseHandle", hwnd); 
 
    /* Free memory. */

    GlobalFree((HGLOBAL)lpBits);
} 
 

HRESULT WINAPI SaveBitmap( const HBITMAP hBmp, LPCTSTR lpszFileName )
{
	try
	{
		PBITMAPINFO pbi = CreateBitmapInfoStruct(NULL, hBmp);

		HDC hDC = GetDC( GetDesktopWindow() );

		CreateBMPFile(
			NULL,
			lpszFileName,
			pbi,
			hBmp,
			hDC );

		ReleaseDC(GetDesktopWindow(), hDC);

		LocalFree(pbi);

		return S_OK;
	}
	catch( HRESULT hr_fail )
	{
		return hr_fail;
	}
	catch(...)
	{
		return E_UNEXPECTED;
	}
}
