//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Hex edit class implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.4 ]
//////////////////////////////////////////////////////////////////////////////////
// HexEdit.cpp : implementation file
//

#include "stdafx.h"
#include "calcem.h"
#include "HexEdit.h"

#define		X_MARGIN	5
#define		Y_MARGIN	3

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CHexEdit

IMPLEMENT_DYNCREATE(CHexEdit, CWnd)

CHexEdit::CHexEdit()
{
//	RegisterWindowClass();
	
	//
	// Assuming 64 kb buffer
	//
	m_cBuffer = 0x10000;
	m_cLines = m_cBuffer >> 4;
	m_CharHeight = 0;

	m_DataMaxPoint.x = 15;

	VERIFY( m_font.CreatePointFont(-110, "Courier New") );
}

CHexEdit::~CHexEdit()
{
}


BEGIN_MESSAGE_MAP(CHexEdit, CWnd)
	//{{AFX_MSG_MAP(CHexEdit)
	ON_WM_CREATE()
	ON_WM_PAINT()
	ON_WM_VSCROLL()
	ON_WM_SIZE()
	ON_WM_SETFOCUS()
	ON_WM_KILLFOCUS()
	ON_WM_KEYDOWN()
	ON_WM_CHAR()
	ON_WM_LBUTTONDOWN()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CHexEdit message handlers


//BOOL CHexEdit::Create(CWnd * pParentWnd, const RECT & rect, DWORD dwStyle, UINT nID)
BOOL CHexEdit::Create(const RECT& rect, CWnd* pParentWnd, UINT nID, DWORD dwStyle)
{
	ASSERT(pParentWnd);
	ASSERT(dwStyle & WS_CHILD);

	return CWnd::CreateEx(
		WS_EX_CLIENTEDGE, 
		AfxRegisterWndClass(
			0,
			AfxGetApp()->LoadStandardCursor(IDC_IBEAM)
			),
		HEXEDIT_CLASSNAME, 
		dwStyle, 
		rect, 
		pParentWnd, 
		nID);
}

int CHexEdit::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	EnableScrollBarCtrl(SB_VERT);

	//
	// Find out char height and width
	//
	// (a) Get DC
	CDC * pDC = GetDC();
	// (b) Set mapping mode 
	pDC->SetMapMode(MM_TEXT);
	// (c) Set font
	CFont * pOldFont = pDC->SelectObject(&m_font);
	// (d) Get text metrics
	TEXTMETRIC tm;
	pDC->GetOutputTextMetrics(&tm);
	// (e) Restore font
	pDC->SelectObject(pOldFont);
	// (f) Release DC
	ReleaseDC(pDC);
	// (g) Save char height & width
	m_CharHeight = tm.tmHeight;		//tm.tmAscent;
	m_CharWidth = tm.tmAveCharWidth;
	
	m_CaretPos.x = X_MARGIN + 6 * m_CharWidth;
	m_CaretPos.y = Y_MARGIN;

	m_Mode = modeHex;

	//
	// Prevent bug occuring during re-use of class
	//
	CRect rect;
	GetClientRect(&rect);
	m_cLinesPerPage = (rect.Height() - 2 * Y_MARGIN) / m_CharHeight;
	SetScrollRange(SB_VERT, 0, m_cLines - m_cLinesPerPage, FALSE);

	return 0;
}

static const char * pszHexLine =
"%04X  %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X  ";

void CHexEdit::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	CFont * pOldFont = dc.SelectObject(&m_font);

	CRect rect;
	GetClientRect(&rect);
	dc.FillSolidRect(&rect, RGB(0xFF,0xFF,0xFF) );

	unsigned offset = GetScrollPos(SB_VERT) << 4;
	char szTemp[72];
	PBYTE pLine = m_pBuffer + offset;
	int y = Y_MARGIN;

	for(unsigned i = 0; i < m_cLinesPerPage; i++, pLine += 16, offset += 16, y += m_CharHeight)
	{
		//
		// Create values
		//
		sprintf(
			szTemp,
			pszHexLine,
			offset,
			pLine[0], pLine[1], pLine[2], pLine[3], pLine[4], pLine[5], pLine[6], pLine[7],
			pLine[8], pLine[9], pLine[10], pLine[11], pLine[12], pLine[13], pLine[14], pLine[15]
			);

		//
		// Create chars
		//
		for(unsigned j = 0; j < 9; j++)
			szTemp[55 + j] = (pLine[j] > 31 && pLine[j] < 127) ? pLine[j] : '.';

		szTemp[55 + j] = ' ';

		for(j = 9; j < 16; j++)
			szTemp[56 + j] = (pLine[j] > 31 && pLine[j] < 127) ? pLine[j] : '.';

		szTemp[56 + j] = 0;
		dc.TextOut(X_MARGIN, y, szTemp);
	}

	//
	// Draw lines between 8th and 9th byte
	//
	int x[2] = 
	{
		X_MARGIN + 29 * m_CharWidth + (m_CharWidth >> 1),
		X_MARGIN + 64 * m_CharWidth + (m_CharWidth >> 1)
	};
	y = 2 * Y_MARGIN + m_cLinesPerPage * m_CharHeight;

	CPen pen(PS_SOLID, 1, 0x808080);
	CPen * pOldPen = dc.SelectObject(&pen);
	for(i = 0; i < 2; i++ )
	{
		dc.MoveTo(x[i], Y_MARGIN);
		dc.LineTo(x[i], y);
	}
	dc.SelectObject(pOldPen);

	dc.SelectObject(pOldFont);
}

#ifdef _DEBUG
static const char * pszScroll[] =
{
	"SB_LINEUP",
	"SB_LINEDOWN",
	"SB_PAGEUP",
	"SB_PAGEDOWN",
	"SB_THUMBPOSITION",
	"SB_THUMBTRACK",
	"SB_TOP",
	"SB_BOTTOM",
	"SB_ENDSCROLL"
};
#endif

void CHexEdit::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
//	TRACE("OnVScroll(%s, %i)\n", pszScroll[nSBCode], (int)nPos);

	switch(nSBCode)
	{
	case SB_TOP: 
		nPos = 0;
		break;

	case SB_BOTTOM: 
		nPos = GetScrollLimit(SB_VERT);
		break;

	case SB_LINEDOWN:
		nPos = GetScrollPos(SB_VERT) + 1;
		break;

	case SB_LINEUP:
		nPos = GetScrollPos(SB_VERT) - 1;
		break;

	case SB_PAGEDOWN:
		nPos = GetScrollPos(SB_VERT) + m_cLinesPerPage;
		break;

	case SB_PAGEUP:
		nPos = GetScrollPos(SB_VERT) - m_cLinesPerPage;
		break;

	case SB_THUMBPOSITION:
		break;

	case SB_THUMBTRACK:
		// Show position somehow
		{
			char szTemp[10];
			sprintf(szTemp, "0x%04X", nPos << 4);
			m_ScrollTip.HandleScrollMessage(this, nSBCode, SB_VERT, szTemp, pScrollBar);
		}
		return;

	case SB_ENDSCROLL:
		// Demand redraw
		InvalidateRect(NULL, FALSE);

		m_ScrollTip.HandleScrollMessage(this, nSBCode, SB_VERT, NULL, pScrollBar);

	default:
		return;
	}

	// Set position
	SetScrollPos(SB_VERT, nPos);

//	CWnd::OnVScroll(nSBCode, nPos, pScrollBar);
}

void CHexEdit::OnSize(UINT nType, int cx, int cy) 
{
	CWnd::OnSize(nType, cx, cy);

	static int last_cy = -1;
	
	if( m_CharHeight && last_cy != cy )
	{
		// Slightly optimized invalidate of bottom part...
		RECT rect;
		GetClientRect(&rect);
		rect.top = rect.bottom - m_CharHeight - Y_MARGIN;
		InvalidateRect(&rect, FALSE);

		last_cy = cy;
		m_cLinesPerPage = (cy - 2 * Y_MARGIN) / m_CharHeight;

		SetScrollRange(SB_VERT, 0, m_cLines - m_cLinesPerPage, FALSE);
//		InvalidateRect(NULL, FALSE);

		m_DataMaxPoint.y = m_cLinesPerPage - 1;
	}
	
}

void CHexEdit::GotoOffset(unsigned offs)
{
	SetScrollPos(SB_VERT, offs >> 4);
}

void CHexEdit::OnSetFocus(CWnd* pOldWnd) 
{
	CWnd::OnSetFocus(pOldWnd);

	// If the user looks at RAM: update on focus!
	if( (GetScrollPos(SB_VERT) + m_cLinesPerPage) > 0x800 )
		InvalidateRect(NULL);

	//
	// Create caret
	//
	::CreateCaret(m_hWnd, NULL, 1, m_CharHeight);
	SetCaretPos(m_CaretPos);
	ShowCaret();
}

void CHexEdit::OnKillFocus(CWnd* pNewWnd) 
{
	CWnd::OnKillFocus(pNewWnd);
	
	//
	// Nuke caret
	//
	DestroyCaret();
}

BOOL CHexEdit::RegisterWindowClass()
{
    WNDCLASS wndcls;
    HINSTANCE hInst = AfxGetInstanceHandle();

    if (!(::GetClassInfo(hInst, HEXEDIT_CLASSNAME, &wndcls)))
    {
        // otherwise we need to register a new class
        wndcls.style            = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
        wndcls.lpfnWndProc      = ::DefWindowProc;
        wndcls.cbClsExtra       = wndcls.cbWndExtra = 0;
        wndcls.hInstance        = hInst;
        wndcls.hIcon            = NULL;
        wndcls.hCursor          = NULL;
        wndcls.hbrBackground    = (HBRUSH) (/*COLOR_3DFACE*/ COLOR_BACKGROUND + 1);
        wndcls.lpszMenuName     = NULL;
        wndcls.lpszClassName    = HEXEDIT_CLASSNAME;

        if (!AfxRegisterClass(&wndcls)) {
            AfxThrowResourceException();
            return FALSE;
        }
    }

    return TRUE;
}

void CHexEdit::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if( m_Mode == modeHexEdit )
	{
		if( nChar == VK_ESCAPE )
		{
			m_Mode = modeHex;
			m_CaretPos.x -= m_CharWidth;
			SetCaretPos(m_CaretPos);
		}

		return;
	}

	switch( nChar )
	{
	case VK_TAB:
		m_Mode = 1 - m_Mode;
		break;

	case VK_PRIOR:	// Page Up
		if( int nPos = GetScrollPos(SB_VERT) )
		{
			SetScrollPos(SB_VERT, nPos - m_cLinesPerPage);
			InvalidateRect(NULL, FALSE);
		}
		break;

	case VK_NEXT:	// Page Down
		SetScrollPos(SB_VERT, GetScrollPos(SB_VERT) + m_cLinesPerPage);
		InvalidateRect(NULL, FALSE);
		break;

	case VK_HOME:
		if( GetKeyState(VK_CONTROL) & 128 )
		{
			// Ctrl+Home
			m_DataPoint.y = 0;
			SetScrollPos(SB_VERT, 0);
			InvalidateRect(NULL, FALSE);
		}

		m_DataPoint.x = 0;
		break;

	case VK_END:
		if( GetKeyState(VK_CONTROL) & 128 )
		{
			// Ctrl+End
			m_DataPoint.y = m_DataMaxPoint.y;
			SetScrollPos(SB_VERT, GetScrollLimit(SB_VERT));
			InvalidateRect(NULL, FALSE);
		}

		m_DataPoint.x = m_DataMaxPoint.x;
		break;

	case VK_LEFT:
		if( m_DataPoint.x > 0 )
		{
			m_DataPoint.x--;
			break;
		}

		if( !m_DataPoint.y )
			break;

		m_DataPoint.x = m_DataMaxPoint.x;
		// fall through

	case VK_UP:
		if( m_DataPoint.y > 0 )
			m_DataPoint.y--;
		else if( int nPos = GetScrollPos(SB_VERT) )
		{
			SetScrollPos(SB_VERT, nPos - 1);
			InvalidateRect(NULL, FALSE);
		}
		break;

	case VK_RIGHT:
		if( m_DataPoint.x < m_DataMaxPoint.x )
		{
			m_DataPoint.x++;
			break;
		}

		m_DataPoint.x = 0;
		// fall through

	case VK_DOWN:
		if( m_DataPoint.y < m_DataMaxPoint.y )
			m_DataPoint.y++;
		else
		{
			SetScrollPos(SB_VERT, GetScrollPos(SB_VERT) + 1);
			InvalidateRect(NULL, FALSE);
		}
		break;

//	default:
//		CWnd::OnKeyDown(nChar, nRepCnt, nFlags);
	}

	UpdateCaretPos();
}


unsigned char ConvertHexChar[0x67] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	} ;


void CHexEdit::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	static BYTE hex_byte;

	if( (m_Mode & modeHex) && isxdigit(nChar) )
	{
		if( m_Mode == modeHex )
		{
			// High nibble
			m_Mode = modeHexEdit;
			hex_byte = ConvertHexChar[nChar] << 4;
		
			// Hide caret and draw nibble
			HideCaret();
			char szTemp[2] = {nChar, 0};
			CDC * pDC = GetDC();
			CFont * pOldFont = pDC->SelectObject(&m_font);
			pDC->TextOut(m_CaretPos.x, m_CaretPos.y, szTemp);
			pDC->SelectObject(pOldFont);
			ReleaseDC(pDC);

			// Advance and show caret
			m_CaretPos.x += m_CharWidth;
			SetCaretPos(m_CaretPos);
			ShowCaret();
		}
		else
		{
			// Low nibble
			m_Mode = modeHex;
			hex_byte |= ConvertHexChar[nChar];
			SetCurrentByte(hex_byte);
		}
	}
	else if( m_Mode == modeChar && nChar > 31)
	{
		SetCurrentByte(nChar);
	}
	
//	CWnd::OnChar(nChar, nRepCnt, nFlags);
}

void CHexEdit::SetCurrentByte(BYTE data)
{
	unsigned offs = ((GetScrollPos(SB_VERT) + m_DataPoint.y) << 4) + m_DataPoint.x;

	m_pBuffer[offs] = data;

	// Somewhat optimized invalidate of row
	RECT rect;
	GetClientRect(&rect);
	rect.top = m_CaretPos.y;
	rect.bottom = rect.top + m_CharHeight; 
	InvalidateRect(&rect, FALSE);

	// Simulate right arrow key press...
	OnKeyDown(VK_RIGHT, 0, 0);
}

void CHexEdit::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// If we're editing, ignore click - TO DO: Cancel editing!!!
	if( m_Mode == modeHexEdit )
		return;
	
	//
	// Convert mouse coordinates to data coordinates
	// (need some improvement)
	//

	// Find and validate line number
	point.y = (point.y - Y_MARGIN) / m_CharHeight;
	if( (unsigned)point.y >= m_cLinesPerPage )
		return;

	m_DataPoint.y = point.y;

	// Convert to half chars
	point.x = (point.x - X_MARGIN) * 2 / m_CharWidth;

	if( point.x < 110 )
	{
		m_Mode = modeHex;

		if( point.x < 12 )
			m_DataPoint.x = 0;

		else if( point.x < 108 )
			m_DataPoint.x = (point.x - 12) / 6;

		else
			m_DataPoint.x = 15;
	}
	else
	{
		m_Mode = modeChar;

		if( point.x < 112 )
			m_DataPoint.x = 0;
		else if( point.x < 128 )
			m_DataPoint.x = (point.x - 110) / 2;
		else if( point.x < 144 )
			m_DataPoint.x = (point.x - 112) / 2;
		else 
			m_DataPoint.x = 15;
	}

		
//	CWnd::OnLButtonDown(nFlags, point);

	UpdateCaretPos();
}

void CHexEdit::UpdateCaretPos()
{
	m_CaretPos.y = Y_MARGIN + m_CharHeight * m_DataPoint.y;

	if( m_Mode == modeHex )
		m_CaretPos.x = X_MARGIN + m_CharWidth * (m_DataPoint.x * 3 + 6);
	else
	{
		int pos = (m_DataPoint.x > 8) ? 56 : 55;
		m_CaretPos.x = X_MARGIN + m_CharWidth * (m_DataPoint.x + pos);
	}

	SetCaretPos(m_CaretPos);
}
