//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Link port helper class implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// LinkPort.cpp: implementation of the CLinkPort class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "calcem.h"
#include "Calculator.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#ifdef _DEBUG
#define ASSERT_T(c)	\
{ if( !(c) ) {		\
	TRACE("Assertion failed at line %i in module " __FILE__ ".\n", __LINE__ );	\
	__asm {int 3}																\
	/* FatalAppExit(0, "Assertion failed!");*/									\
} }
#else
#define ASSERT_T(c)
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkPort::CLinkPort()
{
	m_Dir = Inactive;
	m_pBuffer = NULL;
	m_cBuffer = m_ndxBuffer = 0;
	m_CurrentByte = m_BitsLeft = m_StepsLeft = 0;
}

CLinkPort::~CLinkPort()
{

}


void CLinkPort::LinkOutput()
{
#ifdef _DEBUG
	reg Regs;
	Z80_GetRegs(&Regs);
#endif

	if( m_Dir == Inactive && g_pCalculator->m_LinkState != 3 )
	{
		TRACE("Output to inactive link - start.\n");

		//
		// DANGER - can lock everything...
		//

		// Wait for some change in dir, hopefully to ExpectSend...
		while( m_Dir == Inactive )
			Sleep(0);

		TRACE("Output to inactive link - end (direction = %i).\n", m_Dir);

		ASSERT_T( m_Dir == ExpectSend )
	}

	if( m_Dir == ExpectSend && g_pCalculator->m_LinkState != 3 )
	{
		// Configure
		m_Dir = Send;
		m_BitsLeft = 8;
		m_StepsLeft = 4;
		m_CurrentByte = 0;
	}

	if( m_Dir == Receive )
	{
#ifdef _DEBUG
//		ASSERT_T( m_StepsLeft == 4 || m_StepsLeft == 2 )

		if( m_StepsLeft == 4 )
			ASSERT_T( g_pCalculator->m_LinkState == 1 || g_pCalculator->m_LinkState == 2 )
		else if( m_StepsLeft == 2 )
			ASSERT_T( g_pCalculator->m_LinkState == 3 )
		else
			ASSERT_T(0)
#endif

		m_StepsLeft--;
	}
	
	else if( m_Dir == Send )
	{
#ifdef _DEBUG
//		ASSERT_T( m_StepsLeft == 4 || m_StepsLeft == 2 )

		if( m_StepsLeft == 4 )
			ASSERT_T( g_pCalculator->m_LinkState == 1 || g_pCalculator->m_LinkState == 2 )
		else if( m_StepsLeft == 2 )
			ASSERT_T( g_pCalculator->m_LinkState == 3 )
		else
			ASSERT_T(0)
#endif

		if( m_StepsLeft == 4 )
		{
			ASSERT_T(g_pCalculator->m_LinkState != 3)

			unsigned char bit = g_pCalculator->m_LinkState & 1;
			unsigned char shift = 8 - m_BitsLeft;
			m_CurrentByte |= (bit << shift);
		}

		m_StepsLeft--;
	}
#ifdef _DEBUG
	else if( m_Dir != Inactive )
	{
		TRACE("%04X  Unexpected output %02X (direction = %i).\n", Regs.PC.W, g_pCalculator->m_LinkState, m_Dir);
	}
#endif

}

void CLinkPort::LinkInput()
{
#ifdef _DEBUG
	reg Regs;
	Z80_GetRegs(&Regs);
#endif

	if( m_Dir == ExpectReceive )
	{
		// Retrieve current byte
		m_CurrentByte = m_pBuffer[m_ndxBuffer];

		// Configure
		m_Dir = Receive;
		m_BitsLeft = 8;
		m_StepsLeft = 5;
	}

	if( m_Dir == Receive )
	{
		ASSERT_T( m_StepsLeft == 5 || m_StepsLeft == 3 || m_StepsLeft == 1 )

#ifdef _DEBUG
		if( g_pCalculator->m_Model == 0x83 )
		{
			ASSERT_T( (m_StepsLeft == 5) ? (Regs.PC.W == 0x4e14 || Regs.PC.W == 0x4e2b || Regs.PC.W == 0x4e36) : TRUE );
			ASSERT_T( (m_StepsLeft == 3) ? (Regs.PC.W == 0x4e57 || Regs.PC.W == 0x4e86) : TRUE );
			ASSERT_T( (m_StepsLeft == 1) ? (Regs.PC.W == 0x4e71) : TRUE );
		}
#endif

#ifdef x_DEBUG
		if( g_pCalculator->m_Model == 0x82 )
		{
			ASSERT( (m_StepsLeft == 5) ? (Regs.PC.W == 0x6acb || Regs.PC.W == 0x6ab4 || Regs.PC.W == 0x6ad6) : TRUE );
			ASSERT( (m_StepsLeft == 3) ? (Regs.PC.W == 0x6af8 || Regs.PC.W == 0x6b27) : TRUE );
			ASSERT( (m_StepsLeft == 1) ? (Regs.PC.W == 0x6b12) : TRUE );
		}
#endif

		if( m_StepsLeft == 5 )
		{
			// Next bit..
			g_pCalculator->m_LinkState = (m_CurrentByte & 1) ? 1 : 2;
			m_CurrentByte >>= 1;
		}

		m_StepsLeft--;

		// No In/Out steps left?
		if( !m_StepsLeft )
		{
			// Any bits left?
			if( --m_BitsLeft )
			{
				m_StepsLeft = 5;
			}
			else
			{
				// Any bytes left in buffer?
				if( ++m_ndxBuffer < m_cBuffer )
					m_Dir = ExpectReceive;
				else
					m_Dir = Inactive;
			}
		}
	}
	else if( m_Dir == Send )
	{
		ASSERT_T( m_StepsLeft == 3 || m_StepsLeft == 1 )

		if( m_StepsLeft == 3 )
		{
			// Set state to NULL...
			g_pCalculator->m_LinkState = 0;
		}

		m_StepsLeft--;

		// No In/Out steps left?
		if( !m_StepsLeft )
		{
			// Any bits left?
			if( --m_BitsLeft )
			{
				m_StepsLeft = 4;
			}
			else
			{
				// Save current byte
				m_pBuffer[m_ndxBuffer] = m_CurrentByte;

				// Any space left in buffer?
				if( ++m_ndxBuffer < m_cBuffer )
					m_Dir = ExpectSend;
				else
					m_Dir = Inactive;
			}
		}
	}
#ifdef x_DEBUG
	else if( m_Dir != Inactive )
	{
		TRACE("%04X  Unexpected input (direction = %i).\n", Regs.PC.W, m_Dir);
	}
#endif
}
