//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Main window implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// MainWnd.cpp : implementation file
//

#include "stdafx.h"
#include "calcem.h"
#include "MainWnd.h"
#include "Calculator.h"
#include "AboutDlg.h"
#include "Ti82.h"
#include "Ti83.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////
// CMainWnd

CMainWnd::CMainWnd()
{
}

CMainWnd::~CMainWnd()
{
}

void CMainWnd::EnableIfRunning(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(g_pCalculator != NULL);	
}

void CMainWnd::DisableIfRunning(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(g_pCalculator == NULL);
}


void CMainWnd::OnUpdateFileSendtoemulator(CCmdUI* pCmdUI) 
{
	if( g_pCalculator )
	{
		pCmdUI->Enable(TRUE);
		pCmdUI->SetCheck(g_pCalculator->m_LinkActive ? 1 : 0 );
	}
	else
		pCmdUI->Enable(FALSE);

}	

BEGIN_MESSAGE_MAP(CMainWnd, CWnd)
	//{{AFX_MSG_MAP(CMainWnd)
	ON_WM_CREATE()
	ON_COMMAND(ID_CALC_SCREENDUMP, OnCalcScreendump)
	ON_COMMAND(ID_CALC_START, OnCalcStart)
	ON_COMMAND(ID_CALC_STOP, OnCalcStop)
	ON_COMMAND(ID_DBGWIN_MEMORY, OnDbgwinMemory)
	ON_COMMAND(ID_DBGWIN_REGS, OnDbgwinRegs)
	ON_WM_DESTROY()
	ON_COMMAND(ID_FILE_EMU83_LOAD, OnFileEmu83Load)
	ON_COMMAND(ID_FILE_EMU83_SAVE, OnFileEmu83Save)
	ON_COMMAND(ID_FILE_EXIT, OnFileExit)
	ON_COMMAND(ID_FILE_OPENROM, OnFileOpenrom)
	ON_COMMAND(ID_FILE_OPEN_TI8XEMU, OnFileOpenTi8xemu)
	ON_COMMAND(ID_FILE_SAVE_TI8XEMU, OnFileSaveTi8xemu)
	ON_COMMAND(ID_FILE_SAVE_TI8XEMU2, OnFileSaveTi8xemu2)
	ON_COMMAND(ID_FILE_SENDTOEMULATOR, OnFileSendtoemulator)
	ON_COMMAND(ID_HELP_ABOUT, OnHelpAbout)
	ON_WM_INITMENUPOPUP()
	ON_WM_KILLFOCUS()
	ON_COMMAND(ID_OPT_DISPREFR_IMMED, OnOptDisprefrImmed)
	ON_COMMAND(ID_OPT_DISPREFR_ONINT, OnOptDisprefrOnint)
	ON_WM_PAINT()
	ON_WM_SETFOCUS()
	ON_WM_SIZING()
	ON_UPDATE_COMMAND_UI(ID_FILE_EMU83_SAVE, OnUpdateFileEmu83Save)
	ON_UPDATE_COMMAND_UI(ID_FILE_SENDTOEMULATOR, OnUpdateFileSendtoemulator)
	ON_COMMAND(ID_DBGWIN_ASM, OnDbgwinAsm)
	ON_UPDATE_COMMAND_UI(ID_CALC_SCREENDUMP, EnableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_CALC_START, DisableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_CALC_STOP, EnableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_DBGWIN_MEMORY, EnableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_DBGWIN_REGS, EnableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE_TI8XEMU, EnableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE_TI8XEMU2, EnableIfRunning)
	ON_UPDATE_COMMAND_UI(ID_DBGWIN_ASM, EnableIfRunning)
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
	ON_WM_ENTERMENULOOP()
	ON_WM_EXITMENULOOP()
	ON_COMMAND_EX_RANGE(ID_FILE_MRU_FILE1, ID_FILE_MRU_FILE16, OnOpenRecentFile)
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CMainWnd message handlers

static const char * pszClassName = "CalcEm_MainWindow";

BOOL CMainWnd::Create()
{
	RECT r;
	r.left = AfxGetApp()->GetProfileInt("Main Window", "Left", 100);
	r.top = AfxGetApp()->GetProfileInt("Main Window", "Top", 100);
	r.right = AfxGetApp()->GetProfileInt("Main Window", "Right", 400);
	r.bottom = AfxGetApp()->GetProfileInt("Main Window", "Bottom", 400);

//	RECT rect = {0, 0, 300, 300};	// dummy size & location
	WNDCLASS wndcls;

	wndcls.style			= CS_HREDRAW | CS_VREDRAW;
	wndcls.lpfnWndProc		= ::DefWindowProc;
	wndcls.cbClsExtra		= wndcls.cbWndExtra = 0;
	wndcls.hInstance		= AfxGetInstanceHandle();
	wndcls.hIcon			= AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	wndcls.hCursor			= AfxGetApp()->LoadStandardCursor(IDC_ARROW);
	wndcls.hbrBackground	= 0;
	wndcls.lpszMenuName		= MAKEINTRESOURCE(IDR_CALCEM);
	wndcls.lpszClassName	= pszClassName;

	if( !AfxRegisterClass(&wndcls) )
		AfxThrowResourceException();

	CString strAppTitle;
	strAppTitle.LoadString(AFX_IDS_APP_TITLE);

	return CWnd::CreateEx(
		0, //WS_EX_CLIENTEDGE,
		pszClassName,
		strAppTitle,
		WS_POPUP | WS_VISIBLE | WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX,
		r,
		NULL,
		0 );
}

void CMainWnd::AdjustDlgRect(LPRECT pRect)
{
	if( g_pCalculator )
	{
		// (1) Find out size of borders etc
		unsigned ExtraW = 2 * (GetSystemMetrics(SM_CXSIZEFRAME) + GetSystemMetrics(SM_CXEDGE));
		unsigned ExtraH = 2 * (GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYEDGE))
					  + GetSystemMetrics(SM_CYCAPTION)
					  + GetSystemMetrics(SM_CYMENU)
					  + m_StatusHeight;

		// (2) Find most suitable multiplier
		unsigned size = (pRect->right - pRect->left) / g_pCalculator->m_cxScreen;
		size = max(size, 3);	// multiply with at least 3...
		
		// (3) Set possible width and height
		pRect->right = pRect->left + ExtraW + g_pCalculator->m_cxScreen * size;
		pRect->bottom = pRect->top + ExtraH + g_pCalculator->m_cyScreen * size;
	}
}

int CMainWnd::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;

	m_RefreshOnInt = !!AfxGetApp()->GetProfileInt("Options", "Display Refresh", true);

	// Load last file by default
	CRecentFileList * pRecentFileList = ((CCalcEmApp *)AfxGetApp())->GetRecentFileList();
	if( pRecentFileList )
		m_strRomFile = (*pRecentFileList)[0];

	m_hAccel = ::LoadAccelerators( AfxGetResourceHandle(), MAKEINTRESOURCE(IDR_CALCEM) );

	m_ScreenWnd.CreateEx(
		WS_EX_CLIENTEDGE,
		AfxRegisterWndClass(0, ::LoadCursor(NULL, IDC_ARROW)),
		"Screen",
		WS_CHILD | WS_VISIBLE /*| WS_BORDER*/,
		CRect(0,0,0,0),
		this,
		0x2345 );

	m_StatusBar.Create(
		WS_CHILD | WS_VISIBLE | CCS_BOTTOM | SBARS_SIZEGRIP,
		CRect(0,0,0,0),
		this,
		0x1234 );

	int parts[2] = {150, -1};
	m_StatusBar.SetParts(2, parts);

	m_StatusBar.SetText("Welcome to CalcEm!", 0, 0 );
	m_StatusBar.SetText("", 1, SBT_NOBORDERS );

	CRect status;
	m_StatusBar.GetWindowRect(&status);
	m_StatusHeight = status.Height();

	m_StatusBar.GetRect(0, &status);
	m_Progress.Create(WS_CHILD, status, &m_StatusBar, 0x3456);
	m_Progress.SetRange(0,100);

	return 0;
}

void CMainWnd::OnSizing(UINT fwSide, LPRECT pRect) 
{
	AdjustDlgRect(pRect);

	CWnd::OnSizing(fwSide, pRect);
}

typedef struct
{
	WORD         palVersion; 
	WORD         palNumEntries; 
	PALETTEENTRY palPalEntry[2]; 
} LOGPALETTE1BIT;

void CMainWnd::OnPaint() 
{
	if (IsIconic())
	{
//		PaintIcon();
		return;
	}

	CPaintDC dc(this); // device context for painting

#if 0
	// Get client rect
	CRect rect;
	GetClientRect(&rect);

	rect.bottom -= m_StatusHeight; // note!!!

	if( g_pCalculator )
	{

		// Get source bitmap
		CBitmap * pbmpScreen = g_pCalculator->GetScreenBitmap();

		// Create temporary DC and select bitmap
		CDC dcBmp;
		VERIFY( dcBmp.CreateCompatibleDC(&dc) );
		CBitmap * pbmpOld = dcBmp.SelectObject(pbmpScreen);

		// Stretch and invert bitmap
		dc.StretchBlt(
			0, 0, rect.Width(), rect.Height(),
			&dcBmp,
			0, 0, g_pCalculator->m_cxScreen, g_pCalculator->m_cyScreen,
			NOTSRCCOPY );

		// Clean up
		dcBmp.SelectObject(pbmpOld);
		delete pbmpScreen;

//		dc.SelectPalette(pOldPal, 0);
	}
	else
	{
		dc.FillSolidRect(&rect, 0xffffff);
//		dc.TextOut(10, 10, "Power is off.");

		RECT rect;
		GetClientRect(&rect);

        CDC memDC;
        memDC.CreateCompatibleDC( &dc );

        memDC.SelectObject( &m_LogoBmp );

        // Select and realize the palette
        if( dc.GetDeviceCaps(RASTERCAPS) & RC_PALETTE && m_LogoPal.m_hObject != NULL )
        {
                dc.SelectPalette( &m_LogoPal, FALSE );
                dc.RealizePalette();
        }

		CSize size(75, 74);

        dc.BitBlt(
			(rect.right - size.cx) / 2, 
			(rect.bottom - size.cy) / 2, 
			size.cx, size.cy, 
			&memDC, 0, 0,SRCCOPY);
	}
#endif
	// Do not call CWnd::OnPaint() for painting messages
}

void CMainWnd::OnDestroy() 
{
	RECT r;
	GetWindowRect(&r);
	AfxGetApp()->WriteProfileInt("Main Window", "Left", r.left);
	AfxGetApp()->WriteProfileInt("Main Window", "Top", r.top);
	AfxGetApp()->WriteProfileInt("Main Window", "Right", r.right);
	AfxGetApp()->WriteProfileInt("Main Window", "Bottom", r.bottom);

	CWnd::OnDestroy();
	
	//
	// Stop emulator
	//
	if( g_pCalculator )
		OnCalcStop();
}

void CMainWnd::OnEnterMenuLoop( BOOL bIsTrackPopupMenu )
{
	if( g_pCalculator )
		g_pCalculator->Suspend();
}

void CMainWnd::OnExitMenuLoop( BOOL bIsTrackPopupMenu )
{
	if( g_pCalculator )
		g_pCalculator->Resume();
}

void CMainWnd::OnDbgwinMemory() 
{
	if( m_MemoryDbgWnd.m_hWnd )
		m_MemoryDbgWnd.SetFocus();
	else
		m_MemoryDbgWnd.Create(this, WS_VISIBLE);	
}


void CMainWnd::OnDbgwinRegs() 
{
	if( m_RegsDbgWnd.m_hWnd )
		m_RegsDbgWnd.SetFocus();
	else
		m_RegsDbgWnd.Create(this, WS_VISIBLE);	
}


void CMainWnd::OnFileExit() 
{
	DestroyWindow();	
}

void CMainWnd::OnSetFocus(CWnd* pOldWnd) 
{
	CWnd::OnSetFocus(pOldWnd);
	
	if( g_pCalculator )
	{
//		g_pCalculator->SetInputThread( ::GetCurrentThreadId(), TRUE );

#ifndef x_DEBUG
		g_pCalculator->Resume();
#endif
	}
}

void CMainWnd::OnKillFocus(CWnd* pNewWnd) 
{
	CWnd::OnKillFocus(pNewWnd);

	if( g_pCalculator )
	{
//		g_pCalculator->SetInputThread( ::GetCurrentThreadId(), FALSE );

#ifndef x_DEBUG
		g_pCalculator->Suspend();
#endif
	}
}

void CMainWnd::OnHelpAbout() 
{
	CAboutDlg dlg;
	
	dlg.DoModal();
}

void CMainWnd::OnFileOpenrom() 
{
	CFileDialog dlg(TRUE, "rom", NULL, OFN_HIDEREADONLY, "ROM Images (*.rom)|*.rom|All files (*.*)|*.*||", this);

	if( dlg.DoModal() != IDOK )
		return;

	if( g_pCalculator )
		if( AfxMessageBox(IDS_STOPQ, MB_YESNO) != IDYES )
			return;

	OnCalcStop();
	m_strRomFile = dlg.GetPathName();
}

void CMainWnd::OnCalcStart() 
{
	if( m_strRomFile.IsEmpty() )
	{
		AfxMessageBox(IDS_NOROM);
		return;
	}

	//
	// Start emulator
	//
	if( g_pCalculator )
		OnCalcStop();

	try
	{
		// Detect calculator model
		if( m_strRomFile.Find("82") != -1 )
			g_pCalculator = new CTi82;
		else if( m_strRomFile.Find("83") != -1 )
			g_pCalculator = new CTi83;
		else
			throw (UINT)IDS_ERR_MODEL;

		// Save window handle
		g_pCalculator->m_hwndScreen = m_ScreenWnd.m_hWnd;

		// Load ROM image
		if( !g_pCalculator->LoadROM(m_strRomFile) )
			throw (UINT)IDS_ERR_LOADROM;

		// Reset calculator
		if( !g_pCalculator->Reset() )
			throw (UINT)IDS_ERR_RESET;

		// Load TI8XEMU file?
		if( !m_strRamFile.IsEmpty() )
		{
			if( !g_pCalculator->LoadRAM(m_strRamFile) )
				throw (UINT)IDS_ERR_LOADTI8XEMU;

			m_strRamFile.Empty();
		}

		g_pCalculator->m_RefreshOnInt = m_RefreshOnInt;

		// Start calculator
		if( !g_pCalculator->Start() )
			throw (UINT)IDS_ERR_START;

		// Window title
		CString strModel;
		strModel.Format("TI-%X", g_pCalculator->m_Model);
		m_StatusBar.SetText(strModel, 0, 0);

		// Document
		AfxGetApp()->AddToRecentFileList(m_strRomFile);

		// Menu option
		CheckMenuRadioItem(
			GetMenu()->GetSafeHmenu(),
			ID_OPT_DISPREFR_ONINT,
			ID_OPT_DISPREFR_IMMED,
			ID_OPT_DISPREFR_ONINT + (m_RefreshOnInt ? 0 : 1),
			MF_BYCOMMAND );

		// Update screen
		m_ScreenWnd.OnRefresh();

		// Set input thread
		g_pCalculator->SetInputThread( ::GetCurrentThreadId(), TRUE );
	}
	catch( UINT nIDPrompt )
	{
		// Failure!

		if( g_pCalculator )
		{
			// Delete object
			delete g_pCalculator;
			g_pCalculator = NULL;
		}
	
		// Show message
		AfxMessageBox(nIDPrompt);

		return;
	}

	//
	// Adjust window
	//
	RECT rect;
	GetWindowRect(&rect);
	AdjustDlgRect(&rect);
	MoveWindow(&rect);
//	CenterWindow();
}

void CMainWnd::OnCalcStop() 
{
	if( m_MemoryDbgWnd.m_hWnd )
		m_MemoryDbgWnd.DestroyWindow();

	if( m_RegsDbgWnd.m_hWnd )
		m_RegsDbgWnd.DestroyWindow();

	if( m_AsmDbgWnd.m_hWnd )
		m_AsmDbgWnd.DestroyWindow();

	delete g_pCalculator;
	g_pCalculator = NULL;

	CString strAppTitle;
	strAppTitle.LoadString(AFX_IDS_APP_TITLE);
	
	SetWindowText(strAppTitle);

	m_ScreenWnd.OnRefresh();
}


void CMainWnd::OnInitMenuPopup(CMenu* pMenu, UINT nIndex, BOOL bSysMenu) 
{
//	CWnd::OnInitMenuPopup(pMenu, nIndex, bSysMenu);

	//
	// Successfully ripped from MFC/CFrameWnd/WINFRM.CPP
	//


	ASSERT(pMenu != NULL);
	// check the enabled state of various menu items

	CCmdUI state;
	state.m_pMenu = pMenu;
	ASSERT(state.m_pOther == NULL);
	ASSERT(state.m_pParentMenu == NULL);

	// determine if menu is popup in top-level menu and set m_pOther to
	//  it if so (m_pParentMenu == NULL indicates that it is secondary popup)
	HMENU hParentMenu;
	if (AfxGetThreadState()->m_hTrackingMenu == pMenu->m_hMenu)
		state.m_pParentMenu = pMenu;    // parent == child for tracking popup
	else if ((hParentMenu = ::GetMenu(m_hWnd)) != NULL)
	{
		CWnd* pParent = GetTopLevelParent();
			// child windows don't have menus -- need to go to the top!
		if (pParent != NULL &&
			(hParentMenu = ::GetMenu(pParent->m_hWnd)) != NULL)
		{
			int nIndexMax = ::GetMenuItemCount(hParentMenu);
			for (int nIndex = 0; nIndex < nIndexMax; nIndex++)
			{
				if (::GetSubMenu(hParentMenu, nIndex) == pMenu->m_hMenu)
				{
					// when popup is found, m_pParentMenu is containing menu
					state.m_pParentMenu = CMenu::FromHandle(hParentMenu);
					break;
				}
			}
		}
	}

	state.m_nIndexMax = pMenu->GetMenuItemCount();
	for (state.m_nIndex = 0; state.m_nIndex < state.m_nIndexMax;
	  state.m_nIndex++)
	{
		state.m_nID = pMenu->GetMenuItemID(state.m_nIndex);
		if (state.m_nID == 0)
			continue; // menu separator or invalid cmd - ignore it

		ASSERT(state.m_pOther == NULL);
		ASSERT(state.m_pMenu != NULL);
		if (state.m_nID == (UINT)-1)
		{
			// possibly a popup menu, route to first item of that popup
			state.m_pSubMenu = pMenu->GetSubMenu(state.m_nIndex);
			if (state.m_pSubMenu == NULL ||
				(state.m_nID = state.m_pSubMenu->GetMenuItemID(0)) == 0 ||
				state.m_nID == (UINT)-1)
			{
				continue;       // first item of popup can't be routed to
			}
			state.DoUpdate(this, FALSE);    // popups are never auto disabled
		}
		else
		{
			// normal menu item
			// Auto enable/disable if frame window has 'm_bAutoMenuEnable'
			//    set and command is _not_ a system command.
			state.m_pSubMenu = NULL;
			state.DoUpdate(this, /*m_bAutoMenuEnable &&*/ state.m_nID < 0xF000);
		}

		// adjust for menu deletions and additions
		UINT nCount = pMenu->GetMenuItemCount();
		if (nCount < state.m_nIndexMax)
		{
			state.m_nIndex -= (state.m_nIndexMax - nCount);
			while (state.m_nIndex < nCount &&
				pMenu->GetMenuItemID(state.m_nIndex) == state.m_nID)
			{
				state.m_nIndex++;
			}
		}
		state.m_nIndexMax = nCount;
	}
	
}

void CMainWnd::OnFileOpenTi8xemu() 
{
	CFileDialog dlg(TRUE, NULL, NULL, OFN_HIDEREADONLY, "TI-8x Calculator state|*.82;*.83|All files (*.*)|*.*||", this);

	if( dlg.DoModal() != IDOK )
		return;

	if( g_pCalculator )
		if( AfxMessageBox(IDS_STOPQ, MB_YESNO) != IDYES )
			return;

	OnCalcStop();
	m_strRamFile = dlg.GetPathName();
}

void CMainWnd::OnFileSaveTi8xemu() 
{
	CFileDialog dlg(FALSE, NULL, NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "TI-8x Calculator state|*.82;*.83|All files (*.*)|*.*||", this);

	if( dlg.DoModal() != IDOK )
		return;

	g_pCalculator->StoreRAM( dlg.GetPathName() );
}


void CMainWnd::OnFileSaveTi8xemu2() 
{
	CFileDialog dlg(FALSE, NULL, NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "TI-8x Calculator state|*.82;*.83|All files (*.*)|*.*||", this);

	if( dlg.DoModal() != IDOK )
		return;

	g_pCalculator->StoreRAM( dlg.GetPathName(), 0x0201 );
}

void CMainWnd::OnFileSendtoemulator() 
{
	if( g_pCalculator->m_LinkActive )
	{
		g_pCalculator->StopLink();
	}
	else
	{
		g_pCalculator->Suspend();

		CFileDialog dlg(
			TRUE, NULL, NULL, OFN_HIDEREADONLY | OFN_ALLOWMULTISELECT,
			"TI-8x variables|*.8??|All files (*.*)|*.*||", this);

		int ret = dlg.DoModal();

		g_pCalculator->Resume();

		if( ret != IDOK )
			return;

		CStringArray * pFiles = new CStringArray;
		POSITION p = dlg.GetStartPosition();
		bool BackupFileIncluded = false;
		CString strFile;

		while( p )
		{
			strFile = dlg.GetNextPathName(p);
			pFiles->Add(strFile);
			
			strFile.MakeLower();
			if( strFile.Right(1) == "b" )
				BackupFileIncluded = true;
		}

		if( BackupFileIncluded && pFiles->GetSize() > 1 )
		{
			AfxMessageBox(IDS_ERR_BACKUPMIX);
			return;
		}

		g_pCalculator->SendFiles( pFiles, m_Progress.m_hWnd );
	}
}

// From AFXPRIV.H
#define IS_COMMAND_ID(nID)  ((nID) & 0x8000)

BOOL CMainWnd::OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo) 
{
	//
	// Successfully ripped from MFC/CDialog/DLGCORE.CPP
	//

	if (CWnd::OnCmdMsg(nID, nCode, pExtra, pHandlerInfo))
		return TRUE;

	if ((nCode != CN_COMMAND && nCode != CN_UPDATE_COMMAND_UI) ||
			!IS_COMMAND_ID(nID) || nID >= 0xf000)
	{
		// control notification or non-command button or system command
		return FALSE;       // not routed any further
	}

/*
	// if we have an owner window, give it second crack
	CWnd* pOwner = GetParent();
	if (pOwner != NULL)
	{
#ifdef _DEBUG
		if (afxTraceFlags & traceCmdRouting)
			TRACE1("Routing command id 0x%04X to owner window.\n", nID);
#endif
		ASSERT(pOwner != this);
		if (pOwner->OnCmdMsg(nID, nCode, pExtra, pHandlerInfo))
			return TRUE;
	}
*/

	// last crack goes to the current CWinThread object
	CWinThread* pThread = AfxGetThread();
	if (pThread != NULL)
	{
#ifdef _DEBUG
		if (afxTraceFlags & traceCmdRouting)
			TRACE1("Routing command id 0x%04X to app.\n", nID);
#endif
		if (pThread->OnCmdMsg(nID, nCode, pExtra, pHandlerInfo))
			return TRUE;
	}

#ifdef _DEBUG
	if (afxTraceFlags & traceCmdRouting)
	{
		TRACE2("IGNORING command id 0x%04X sent to %hs main window.\n", nID,
				GetRuntimeClass()->m_lpszClassName);
	}
#endif
	return FALSE;
}

BOOL CMainWnd::OnOpenRecentFile(UINT nID)
{
	if( g_pCalculator )
		if( AfxMessageBox(IDS_STOPQ, MB_YESNO) != IDYES )
			return TRUE;	// prevent further dispatch!
	
	//
	// Successfully ripped from MFC
	//

	CRecentFileList * pRecentFileList = ((CCalcEmApp *)AfxGetApp())->GetRecentFileList();

	ASSERT_VALID(this);
	ASSERT(pRecentFileList != NULL);

	ASSERT(nID >= ID_FILE_MRU_FILE1);
	ASSERT(nID < ID_FILE_MRU_FILE1 + (UINT)pRecentFileList->GetSize());
	int nIndex = nID - ID_FILE_MRU_FILE1;
	ASSERT((*pRecentFileList)[nIndex].GetLength() != 0);

	TRACE2("MRU: open file (%d) '%s'.\n", (nIndex) + 1,
			(LPCTSTR)(*pRecentFileList)[nIndex]);

	m_strRomFile = (*pRecentFileList)[nIndex];

	pRecentFileList->Remove(nIndex);	// added again if it works...
	OnCalcStart();

//	if (OpenDocumentFile((*pRecentFileList)[nIndex]) == NULL)
//		pRecentFileList->Remove(nIndex);

	return TRUE;	// prevent further dispatch!
}

void CMainWnd::OnOptDisprefrOnint() 
{
	if( g_pCalculator )
	{
		m_RefreshOnInt = g_pCalculator->m_RefreshOnInt = true;

		CheckMenuRadioItem(
			GetMenu()->GetSafeHmenu(),
			ID_OPT_DISPREFR_ONINT,
			ID_OPT_DISPREFR_IMMED,
			ID_OPT_DISPREFR_ONINT,
			MF_BYCOMMAND );
	}
}

void CMainWnd::OnOptDisprefrImmed() 
{
	if( g_pCalculator )
	{
		m_RefreshOnInt = g_pCalculator->m_RefreshOnInt = false;

		CheckMenuRadioItem(
			GetMenu()->GetSafeHmenu(),
			ID_OPT_DISPREFR_ONINT,
			ID_OPT_DISPREFR_IMMED,
			ID_OPT_DISPREFR_IMMED,
			MF_BYCOMMAND );
	}
}

void CMainWnd::OnFileEmu83Load() 
{
	CFileDialog dlg(TRUE, NULL, NULL, OFN_HIDEREADONLY, "EMU83 State image (.img)|*.img|All files (*.*)|*.*||", this);

	if( dlg.DoModal() != IDOK )
		return;

	if( g_pCalculator )
		if( AfxMessageBox(IDS_STOPQ, MB_YESNO) != IDYES )
			return;

	OnCalcStop();
	m_strRamFile = dlg.GetPathName();
}

void CMainWnd::OnFileEmu83Save() 
{
	CFileDialog dlg(FALSE, NULL, NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "EMU83 State image (.img)|*.img|All files (*.*)|*.*||", this);

	if( dlg.DoModal() != IDOK )
		return;

	g_pCalculator->StoreRAM( dlg.GetPathName() );
}

void CMainWnd::OnUpdateFileEmu83Save(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(g_pCalculator ? g_pCalculator->m_Model == 0x83 : FALSE);	
}

void CMainWnd::OnCalcScreendump() 
{
	g_pCalculator->Suspend();
	CBitmap * pbmpScreen = g_pCalculator->GetScreenBitmap();

	CFileDialog dlg(FALSE, "bmp", NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, "Bitmap (.bmp)|*.bmp|All files (*.*)|*.*||", this);

	if( dlg.DoModal() == IDOK )
		SaveBitmap( *pbmpScreen, dlg.GetPathName() );

	delete pbmpScreen;
	g_pCalculator->Resume();
}

BOOL CMainWnd::PreTranslateMessage(MSG* pMsg) 
{
	if (CWnd::PreTranslateMessage(pMsg))
		return TRUE;
	
	if (pMsg->message >= WM_KEYFIRST && pMsg->message <= WM_KEYLAST)
	{
		return ::TranslateAccelerator(m_hWnd, m_hAccel, pMsg);
	}
	
	return FALSE;
}

void CMainWnd::OnDbgwinAsm() 
{
	if( m_AsmDbgWnd.m_hWnd )
		m_AsmDbgWnd.SetFocus();
	else
		m_AsmDbgWnd.Create(this, WS_VISIBLE);	
}

void CMainWnd::OnSize(UINT nType, int cx, int cy) 
{
	CWnd::OnSize(nType, cx, cy);
	
	m_ScreenWnd.MoveWindow(0, 0, cx, cy - m_StatusHeight);
	m_StatusBar.MoveWindow(0, cy - m_StatusHeight, cx, m_StatusHeight);
}
