//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ CMiniCalculator class implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////
// MiniCalculator.cpp: implementation of the CMiniCalculator class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "calcem.h"
#include "MiniCalculator.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
 
CMiniCalculator::CMiniCalculator()
{
	m_Model = 82;
	m_cxScreen = 96;
	m_cyScreen = 64;

	m_pMemory = new unsigned char[0x10000];					// 64 kb
	m_pScreen = new BYTE[768];
}

CMiniCalculator::~CMiniCalculator()
{

}

//----------------------------------------------------------------------------
//
// Display Control
//
//----------------------------------------------------------------------------
void CMiniCalculator::DisplayControl(unsigned char Cmd)
{
	if( Cmd <= 1 )
	{
		//
		// Display mode
		//
		m_DisplayMode = Cmd;
	}
	else if( Cmd >= 4 && Cmd <= 7 )
	{
		//
		// Up, Down, Left, Right
		//
		m_DisplayMove = Cmd;
	}
	else if( (Cmd & 0xC0) == 0x40 )
	{
		//
		// set hardware scroll
		//
#ifdef x_DEBUG
		if( Cmd & 0x3F )
			TRACE("DisplayControl: Hardware scroll = %i lines.\n", Cmd & 0x3F);
#endif
		//
		// TO DO: Implement it!
		//
	}
	else if( (Cmd & 0xE0) == 0x20 )
	{
		//
		// set column
		//
		m_DispX = Cmd & 0x1F;
//		ASSERT(m_DispX < 0xC);
		m_CursorMoved = true;
	}
	else if( (Cmd & 0xC0) == 0x80 )
	{
		//
		// set line
		//
		m_DispY = Cmd & 0x3F;
		m_CursorMoved = true;
	}

#ifdef _DEBUG
	else if( (Cmd & 0xC0) == 0xC0 )
	{
		TRACE("DisplayControl: Contrast = %i\n", Cmd & 0x3F);
	}

	else if( Cmd == 2 )
		TRACE("DisplayControl: Stop\n");
	else if( Cmd == 3 )
		TRACE("DisplayControl: Start\n");
	else
		TRACE("DisplayControl: Unknown %02X.\n", Cmd);
#endif
}

//----------------------------------------------------------------------------
//
// Display Output
//
//----------------------------------------------------------------------------
void CMiniCalculator::DisplayOutput(unsigned char Data)
{
	ASSERT( m_DispY < m_cyScreen );

	// Not good!!!
	if( m_DispY >= m_cyScreen )
		return;

	if( m_DisplayMode )
	{
/*
		ASSERT( m_DispX < (m_cxScreen >> 3) );

		// Not good!!!
		if( m_DispX >= (m_cxScreen >> 3) )
			return;
*/

		unsigned offset = m_DispY * 12 + m_DispX;
//		ASSERT(offset < 0x300);
		
		if( offset < 0x300 )
			m_pScreen[ offset ] = Data;
	}
	else
	{
		//
		// Low three bits of column gives the bit shift
		// (the number of bits in second byte used on screen)
		// High bits gives the byte offset on line
		//
	
		unsigned column = 6 * m_DispX;

/*
		ASSERT( (column >> 3) < (m_cxScreen >> 3) );
		// Not good!!!
		if( (column >> 3) >= (m_cxScreen >> 3) )
			return;
*/
		unsigned offset = m_DispY * 12 + (column >> 3);
//		ASSERT(offset < 0x300);

		if( offset < 0x300 )
		{
			unsigned char shift = column & 7;
			unsigned char mask;

			Data <<= 2;

			mask = ~(252 >> shift);
			m_pScreen[offset] = m_pScreen[offset] & mask | (Data >> shift);

			if( shift > 2 && offset < 0x2ff )
			{
				offset++;
				
				shift = 8 - shift;

				mask = ~(252 << shift);
				m_pScreen[offset] = m_pScreen[offset] & mask | (Data << shift);
			}
		}
	}
		
	// Up, Down, Left, Right
	switch( m_DisplayMove )
	{
	case 4:
		m_DispY--;
		m_DispY &= 0x3f;
		break;
	case 5:
		m_DispY++;
		m_DispY &= 0x3f;
		break;
	case 6:
		m_DispX--;
		m_DispX &= 0xf;
		break;
	case 7:
		m_DispX++;
		m_DispX &= 0xf;
		break;
	}

	// Update screen
	if( !m_RefreshOnInt )
	{
		// Screen is updated, tell Windows to redraw
		ASSERT(m_hwndScreen);
		::PostMessage(m_hwndScreen, wm_Refresh, 0, 0);
	}
}

//----------------------------------------------------------------------------
//
// Display Input
//
//----------------------------------------------------------------------------
unsigned char CMiniCalculator::DisplayInput()
{
	// Weird but just as in 82-ports.txt
	if( m_CursorMoved )
	{
		m_CursorMoved = false;
		return 0;
	}

	byte Data;

	if( m_DisplayMode )
	{
		ASSERT( m_DispX < (m_cxScreen >> 3) );
		Data = m_pScreen[ m_DispY * 12 + m_DispX ];
	}
	else
	{
		unsigned column = 6 * m_DispX;
		unsigned offset = m_DispY * 12 + (column >> 3);

		unsigned char shift = 10 - (column & 7);

		Data = ((m_pScreen[offset] << 8) | m_pScreen[offset + 1]) >> shift;
	}

	// Up, Down, Left, Right
	switch( m_DisplayMove )
	{
	case 4:
		m_DispY--;
		m_DispY &= 0x3f;
		break;
	case 5:
		m_DispY++;
		m_DispY &= 0x3f;
		break;
	case 6:
		m_DispX--;
		m_DispX &= 0xf;
		break;
	case 7:
		m_DispX++;
		m_DispX &= 0xf;
		break;
	}

	return Data;
}

/*

Structure of stored registers...

typedef struct
{
  pair AF,BC,DE,HL,IX,IY,PC,SP;
  pair AF2,BC2,DE2,HL2;
  unsigned IFF1,IFF2,HALT,IM,I,R,R2;
} Z80_Regs;

*/

/*
bool CMiniCalculator::LoadTi8xEmuRAM(const char * pszFile)
{
	return false;
#if 0
	CWaitCursor cur;
	CFile file(pszFile, CFile::modeRead);
	
	

	if( file.GetLength() == 0x832B )
	{
		//
		// TI-82 saved image from TI-8X emulator v0.2b1
		//
	}
	else if( file.GetLength() == 0x8327 )
	{
		//
		// TI-82 saved image from TI-8X emulator v0.1
		//
	}
	
	

	if( file.GetLength() != 0x8327 )
		return false;

	unsigned char temp[0x27];

	// RAM image
	file.Read(m_pMemory + 0x8000, 0x8000);

	// Misc
	file.Read(temp, 0x27);

		// These register pairs are stored "backwards"
		// AF, BC, DE, HL, IX, IY, PC, SP, AF', BC', DE', HL'
		_swab((char *)temp, (char *)&m_Regs, 0x18);
		//
		// You don't know what this function does? Read the fine manual:
		//
		// void _swab( char *src, char *dest, int n );
		//
		// The _swab function copies n bytes from src, swaps each pair of
		// adjacent bytes, and stores the result at dest. The integer n
		// should be an even number to allow for swapping. _swab is typically 
		// used to prepare binary data for transfer to a machine that uses a 
		// different byte order.
		//

/x*
		memcpy(&m_Regs, temp, 24 );

		void * pRegs = &m_Regs;

		__asm
		{
			mov		ecx, 12
			lea		eax, pRegs
			mov		ebx, [eax]

		next_reg:
			ror		word ptr [ebx], 8
			inc		ebx
			inc		ebx

			loop	next_reg	
		}
*x/

		// IFF1
		if( temp[0x18] )
			m_Regs.IFF |= 1;

		// IFF2
		if( temp[0x19] )
			m_Regs.IFF |= 64;

		// HALT
		if( temp[0x1A] )
			m_Regs.IFF |= 128;

		// Interrupt mode (0 to 2)
//		m_Regs.IFF |= (temp[0x1B] << 1);

		// I
		m_Regs.I = temp[0x1C];

		// R, R2: two bytes

		// Keyboard mask
		m_KeyboardMask = temp[0x1F];

		// ON mask
		m_MaskON = temp[0x20];

		// Unknown/unused: one byte - WEIRD!!

		// ROM page
		m_RomPage = temp[0x22];

		// Display stuff
		m_DispX = temp[0x23];
		m_DispY = temp[0x24];
		m_DisplayMove = temp[0x25];
		m_DisplayMode = temp[0x26];

	// Screen image
	file.Read(m_pScreen, 768);

	return true;
#endif
}

bool CMiniCalculator::StoreTi8xEmuRAM( const char * pszFile, unsigned nVersion )
{
	CWaitCursor cur;

	// Stop execution at next interrupt...
	::ResetEvent(m_hInterrupt);
	::ResetEvent(m_hInterruptNoWait);
	::WaitForSingleObject(m_hInterrupt, INFINITE);
	
	GetUpToDateRegs();

	CFile file(pszFile, CFile::modeCreate | CFile::modeWrite );

	unsigned char temp[0x27];

	// RAM image
	file.Write(m_pMemory + 0x8000, 0x8000);

		// These register pairs are stored "backwards"
		// AF, BC, DE, HL, IX, IY, PC, SP, AF', BC', DE', HL'
		_swab((char *)&m_Regs, (char *)temp, 24);
		//
		// You don't know what this function does? Read the fine manual:
		//
		// void _swab( char *src, char *dest, int n );
		//
		// The _swab function copies n bytes from src, swaps each pair of
		// adjacent bytes, and stores the result at dest. The integer n
		// should be an even number to allow for swapping. _swab is typically 
		// used to prepare binary data for transfer to a machine that uses a 
		// different byte order.
		//

/x*
		memcpy(temp, &m_Regs, 24 );

		void * pRegs = temp;

		__asm
		{
			mov		ecx, 12
			lea		eax, pRegs
			mov		ebx, [eax]

		next_reg:
			ror		word ptr [ebx], 8
			inc		ebx
			inc		ebx

			loop	next_reg	
		}
*x/
		// IFF1
		temp[0x18] = !!(m_Regs.IFF & 1);

		// IFF2
		temp[0x19] = !!(m_Regs.IFF & 64);

		// HALT
		temp[0x1A] = !!(m_Regs.IFF & 128);

		// Interrupt mode (0 to 2)
//		m_Regs.IFF |= (temp[0x1B] << 1);
		temp[0x1B] = 0;		// ????

		// I
		temp[0x1C] = m_Regs.I;

		// R, R2: two bytes
		temp[0x1D] = temp[0x1E] = 0;

		// Keyboard mask
		temp[0x1F] = m_KeyboardMask;

		// ON mask
		temp[0x20] = m_MaskON;

		// Contrast
		temp[0x21] = 0;

		if( m_Model == 0x82 )
		{
			// ROM page
			temp[0x22] = m_RomPage;

			// Display stuff
			temp[0x23] = m_DispX;
			temp[0x24] = m_DispY;
			temp[0x25] = m_DisplayMove;
			temp[0x26] = m_DisplayMode;
			temp[0x27] = 0;					// Hardware scroll
		}
		else if( m_Model == 0x83 )
		{
			// ROM page
			temp[0x22] = m_RomPage;

			// Display stuff
			temp[0x23] = m_DispX;
			temp[0x24] = m_DispY;
			temp[0x25] = m_DisplayMove;
			temp[0x26] = m_DisplayMode;
		}
		


	// Misc
	file.Write(temp, 0x27);

	// Screen image
	file.Write(m_pScreen, 768);


	// Continue execution
	::SetEvent(m_hInterruptNoWait);

	return true;
}
*/

bool CMiniCalculator::Reset()
{
	if( !CCalculator::Reset() )
		return false;

	// Internal stuff
	m_KeyboardMask = 0;
	m_MaskON = 0;
	m_ONKeyDown = 0;
//	m_DisplayUpdated = false;
	m_DisplayMode = 0;
	m_DisplayMove = 7;
	m_CursorMoved = false;

	return true;
}
