//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Screen window implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// ScreenWnd.cpp : implementation file
//

#include "stdafx.h"
#include "calcem.h"
#include "ScreenWnd.h"
#include "Calculator.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CScreenWnd

CScreenWnd::CScreenWnd()
{
}

CScreenWnd::~CScreenWnd()
{
}


BEGIN_MESSAGE_MAP(CScreenWnd, CWnd)
	//{{AFX_MSG_MAP(CScreenWnd)
	ON_WM_PAINT()
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
	ON_REGISTERED_MESSAGE(wm_Refresh, OnRefresh)
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CScreenWnd message handlers


LONG CScreenWnd::OnRefresh( UINT wParam, LONG lParam )
{
//	RECT rect;
//	GetClientRect(&rect);
//	rect.bottom -= m_StatusHeight;
//	InvalidateRect(&rect);
	InvalidateRect(NULL);

	return 0;	// what return value should be given???
}
void CScreenWnd::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	// Get client rect
	CRect rect;
	GetClientRect(&rect);

	if( g_pCalculator )
	{

		// Get source bitmap
		CBitmap * pbmpScreen = g_pCalculator->GetScreenBitmap();

		// Create temporary DC and select bitmap
		CDC dcBmp;
		VERIFY( dcBmp.CreateCompatibleDC(&dc) );
		CBitmap * pbmpOld = dcBmp.SelectObject(pbmpScreen);

		// Stretch and invert bitmap
		dc.StretchBlt(
			0, 0, rect.Width(), rect.Height(),
			&dcBmp,
			0, 0, g_pCalculator->m_cxScreen, g_pCalculator->m_cyScreen,
			NOTSRCCOPY );

		// Clean up
		dcBmp.SelectObject(pbmpOld);
		delete pbmpScreen;

//		dc.SelectPalette(pOldPal, 0);
	}
	else
	{
		dc.FillSolidRect(&rect, 0xffffff);
//		dc.TextOut(10, 10, "Power is off.");

		RECT rect;
		GetClientRect(&rect);

        CDC memDC;
        memDC.CreateCompatibleDC( &dc );

        memDC.SelectObject( &m_LogoBmp );

        // Select and realize the palette
        if( dc.GetDeviceCaps(RASTERCAPS) & RC_PALETTE && m_LogoPal.m_hObject != NULL )
        {
                dc.SelectPalette( &m_LogoPal, FALSE );
                dc.RealizePalette();
        }

		CSize size(75, 74);

        dc.BitBlt(
			(rect.right - size.cx) / 2, 
			(rect.bottom - size.cy) / 2, 
			size.cx, size.cy, 
			&memDC, 0, 0,SRCCOPY);
	}
	
	// Do not call CWnd::OnPaint() for painting messages
}

BOOL GetBitmapAndPalette(UINT nIDResource, CBitmap &bitmap, CPalette &pal)
{
        LPCTSTR lpszResourceName = (LPCTSTR)nIDResource;

        HBITMAP hBmp = (HBITMAP)::LoadImage( AfxGetInstanceHandle(), 
                        lpszResourceName, IMAGE_BITMAP, 0,0, LR_CREATEDIBSECTION );

        if( hBmp == NULL ) 
                return FALSE;

        bitmap.Attach( hBmp );

        // Create a logical palette for the bitmap
        DIBSECTION ds;
        BITMAPINFOHEADER &bmInfo = ds.dsBmih;
        bitmap.GetObject( sizeof(ds), &ds );

        int nColors = bmInfo.biClrUsed ? bmInfo.biClrUsed : 1 << bmInfo.biBitCount;

        // Create a halftone palette if colors > 256. 
        CClientDC dc(NULL);                     // Desktop DC
        if( nColors > 256 )
                pal.CreateHalftonePalette( &dc );
        else
        {
                // Create the palette

                RGBQUAD *pRGB = new RGBQUAD[nColors];
                CDC memDC;
                memDC.CreateCompatibleDC(&dc);

                memDC.SelectObject( &bitmap );
                ::GetDIBColorTable( memDC, 0, nColors, pRGB );

                UINT nSize = sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * nColors);
                LOGPALETTE *pLP = (LOGPALETTE *) new BYTE[nSize];

                pLP->palVersion = 0x300;
                pLP->palNumEntries = nColors;

                for( int i=0; i < nColors; i++)
                {
                        pLP->palPalEntry[i].peRed = pRGB[i].rgbRed;
                        pLP->palPalEntry[i].peGreen = pRGB[i].rgbGreen;
                        pLP->palPalEntry[i].peBlue = pRGB[i].rgbBlue;
                        pLP->palPalEntry[i].peFlags = 0;
                }

                pal.CreatePalette( pLP );

                delete[] pLP;
                delete[] pRGB;
        }

        return TRUE;
}

int CScreenWnd::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	GetBitmapAndPalette(IDB_CALCEM, m_LogoBmp, m_LogoPal);
	
	return 0;
}
