//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Disassembly window implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.4 ]
//////////////////////////////////////////////////////////////////////////////////
// AsmWnd.cpp: implementation of the CAsmWnd class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "calcem.h"
#include "Calculator.h"
#include "AsmWnd.h"

extern "C"
{
	int DAsm(char *S,unsigned short A);
}

#define		X_MARGIN	5
#define		Y_MARGIN	3

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BEGIN_MESSAGE_MAP(CAsmWnd, CFloatWnd)
	//{{AFX_MSG_MAP(CAsmWnd)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_PAINT()
	ON_WM_VSCROLL()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CAsmWnd::CAsmWnd()
{
	VERIFY( m_font.CreatePointFont(-110, "Courier New") );

	m_BytesOnPage = m_Offset = m_cLinesPerPage = m_CharHeight = m_CharWidth = 0;
}

CAsmWnd::~CAsmWnd()
{

}

BOOL CAsmWnd::Create(CWnd * pParent, DWORD dwStyle)
{
	RECT r;
	r.left = AfxGetApp()->GetProfileInt("Asm Window", "Left", 50);
	r.top = AfxGetApp()->GetProfileInt("Asm Window", "Top", 50);
	r.right = AfxGetApp()->GetProfileInt("Asm Window", "Right", 700);
	r.bottom = AfxGetApp()->GetProfileInt("Asm Window", "Bottom", 250);

	return CFloatWnd::Create(
		"Disassembly",
		r,
		pParent,
		dwStyle | WS_THICKFRAME );
}

//#define NO_OWNERDRAW

int CAsmWnd::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CFloatWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	EnableScrollBar(SB_VERT);

	//
	// Find out char height and width
	//
	// (a) Get DC
	CDC * pDC = GetDC();
	// (b) Set mapping mode 
	pDC->SetMapMode(MM_TEXT);
	// (c) Set font
	CFont * pOldFont = pDC->SelectObject(&m_font);
	// (d) Get text metrics
	TEXTMETRIC tm;
	pDC->GetOutputTextMetrics(&tm);
	// (e) Restore font
	pDC->SelectObject(pOldFont);
	// (f) Release DC
	ReleaseDC(pDC);
	// (g) Save char height & width
	m_CharHeight = tm.tmHeight;		//tm.tmAscent;
	m_CharWidth = tm.tmAveCharWidth;

	//
	// Stuff
	//
	CRect rect;
	GetClientRect(&rect);
	m_cLinesPerPage = (rect.Height() - 2 * Y_MARGIN) / m_CharHeight;
	SetScrollRange(SB_VERT, 0, 0x3fff, FALSE);

#if 0

	RECT r;
	GetClientRect(&r);

	m_ListBox.Create(
		WS_VISIBLE | WS_CHILD | WS_TABSTOP |
#ifndef NO_OWNERDRAW
		LBS_HASSTRINGS | LBS_OWNERDRAWFIXED | 
#endif
		LBS_USETABSTOPS | LBS_NOINTEGRALHEIGHT,
		r,
		this,
		0xde );

	m_ListBox.AddString("0000  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0001  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0002  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0003  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0004  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0005  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0006  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0007  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0008  00 00 00 00  call  z, nisse");
	m_ListBox.AddString("0009  00 00 00 00  call  z, nisse");
#endif


	return 0;
}

void CAsmWnd::OnDestroy() 
{
	CFloatWnd::OnDestroy();
	
	RECT r;
	GetWindowRect(&r);
	AfxGetApp()->WriteProfileInt("Asm Window", "Left", r.left);
	AfxGetApp()->WriteProfileInt("Asm Window", "Top", r.top);
	AfxGetApp()->WriteProfileInt("Asm Window", "Right", r.right);
	AfxGetApp()->WriteProfileInt("Asm Window", "Bottom", r.bottom);
}

void CAsmWnd::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	
	CFont * pOldFont = dc.SelectObject(&m_font);

	CRect rect;
	GetClientRect(&rect);
	dc.FillSolidRect(&rect, g_bgWindow );

	char szDAsm[64];
	char szData[16];
	char * pszData;;
	char szTemp[128];
	int size = 1;
	unsigned offset = m_Offset;
	int y = Y_MARGIN;

	for(unsigned i = 0; i < m_cLinesPerPage; i++, y += m_CharHeight)
	{
		size = DAsm(szDAsm, offset);

		if( size > 0 )
		{
			pszData = szData;

			for(int j = 0; j < size; j++)
			{
				sprintf(pszData, "%02X ", g_pCalculator->m_pMemory[offset + j]);
				pszData += 3;
			}

			sprintf(
				szTemp,
				"%04X  %-13s%s",
				offset,
				szData,
				szDAsm );
		}
		else
		{
			if( offset != 0x10000 )
			{
				size = 0x10000 - offset;

				sprintf(
					szTemp,
					"%04X  %s",
					offset,
					"(bytes)" );
			}

			i = 0xffffffff;
		}

		dc.TextOut(X_MARGIN, y, szTemp);

		offset += size;
	}

	m_BytesOnPage = offset - m_Offset;

	dc.SelectObject(pOldFont);
}

void CAsmWnd::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	int nOffs = (int)m_Offset;

	switch(nSBCode)
	{
	case SB_TOP: 
		nOffs = 0;
		break;

	case SB_BOTTOM: 
		nOffs = 0xffff;
		break;

	case SB_LINEDOWN:
		nOffs++;
		break;

	case SB_LINEUP:
		nOffs--;
		break;

	case SB_PAGEDOWN:
		nOffs += m_BytesOnPage;
		break;

	case SB_PAGEUP:
		nOffs -= m_BytesOnPage;
		break;

	case SB_THUMBPOSITION:
		nOffs = nPos << 2;
		break;

	case SB_THUMBTRACK:
		// Show position somehow
		{
			char szTemp[10];
			sprintf(szTemp, "0x%04X", nPos << 2);
			m_ScrollTip.HandleScrollMessage(this, nSBCode, SB_VERT, szTemp, pScrollBar);
		}
		return;

	case SB_ENDSCROLL:
		// Demand redraw
		InvalidateRect(NULL, FALSE);

		m_ScrollTip.HandleScrollMessage(this, nSBCode, SB_VERT, NULL, pScrollBar);

		// fall through

	default:
		return;
	}

	if( nOffs >= 0x10000 )
		nOffs = 0xffff;
	else if( nOffs < 0 )
		nOffs = 0;

	m_Offset = (unsigned)nOffs;
	nPos = m_Offset >> 2;

	// Set position
	SetScrollPos(SB_VERT, nPos );
	
//	CFloatWnd::OnVScroll(nSBCode, nPos, pScrollBar);
}

void CAsmWnd::OnSize(UINT nType, int cx, int cy) 
{
	CFloatWnd::OnSize(nType, cx, cy);
	
	static int last_cy = -1;
	
	if( m_CharHeight && last_cy != cy )
	{
		// Slightly optimized invalidate of bottom part...
		RECT rect;
		GetClientRect(&rect);
		rect.top = rect.bottom - m_CharHeight - Y_MARGIN;
		InvalidateRect(&rect, FALSE);

		last_cy = cy;
		m_cLinesPerPage = (cy - 2 * Y_MARGIN) / m_CharHeight;
	}
}
