//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ Linking functions implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "Link.h"
#include "Calculator.h"

static void ReadBuffer(unsigned char * pBuffer, unsigned cBuffer);
static void WriteBuffer(unsigned char * pBuffer, unsigned cBuffer);

static WORD CalculateLinkDataChecksum(PBYTE pData, WORD wLength)
{
	WORD Checksum = 0;

	for( DWORD i = 0; i < wLength; i++)
		Checksum += pData[i];

	return Checksum;
}

typedef struct 
{
	BYTE Calculator;
	BYTE Command;
	WORD Length;
} TIPACKAGEHEADER, *PTIPACKAGEHEADER;

//----------------------------------------------------------------------------
//
// Allow the Emulator to receive
//
//----------------------------------------------------------------------------

unsigned char LinkCommand[2][4] =
{
	{0x06, 0x15, 0x06, 0x15 },
	{0x06, 0x15, 0x15, 0x15 },

//	{0xc9, 0x15, 0xc9, 0x15 },
//	{0xc9, 0x15, 0x15, 0x15 },
};

#define BLOCK_LEN		1024

bool SendFile(unsigned Type, const char * pszFile, CProgressCtrl * pProgress)
{
	int Progress = 0;

	//
	// Read file
	//
	CFile file(pszFile, CFile::modeRead);
	file.Seek(0x35, CFile::begin);

	unsigned short totalSize;
	file.Read(&totalSize, 2);

	unsigned char * pBuffer = new unsigned char[totalSize];
	if( file.Read(pBuffer, totalSize) != totalSize )
	{
		TRACE("Invalid file!\n");
		return 1;
	}

	file.Close();

	//
	// Prepare transfer
	//
	unsigned short wChecksum, size;
	unsigned char * pTemp = pBuffer;
	TIPACKAGEHEADER tiph;
	unsigned nBlock = 0;

	pProgress->SetRange(0, totalSize);

	while( totalSize )
	{
		//
		// Send block
		//
		size = *(unsigned short *)pTemp;
		pTemp += 2;

		tiph.Calculator = g_pCalculator->m_Model & 0xF;
		tiph.Command = LinkCommand[Type][nBlock & 3];
		tiph.Length = size;

		wChecksum = CalculateLinkDataChecksum(pTemp, size);

		WriteBuffer((unsigned char *)&tiph, 4);

		while( size > BLOCK_LEN )
		{
			WriteBuffer(pTemp, BLOCK_LEN);	

			pTemp += BLOCK_LEN;
			Progress += BLOCK_LEN;

			size -= BLOCK_LEN;
			totalSize -= BLOCK_LEN;
			pProgress->SetPos(Progress);
		}

		WriteBuffer(pTemp, size);
		WriteBuffer((unsigned char *)&wChecksum, 2);
		ReadBuffer((unsigned char *)&tiph, 4);

		if( tiph.Command != 0x56 )
			throw 1;

		if( (Type == 0 && (nBlock & 1) == 0) || (Type == 1 && nBlock == 0) )
		{
			ReadBuffer((unsigned char *)&tiph, 4);

			if( tiph.Command != 0x09 )
				throw 3;

			tiph.Calculator = g_pCalculator->m_Model & 0xF;
			tiph.Command = 0x56;
			tiph.Length = 0;

			WriteBuffer((unsigned char *)&tiph, 4);
		}

		Progress += size + 2;
		pTemp += size;
		totalSize -= (size + 2);
		nBlock++;

		pProgress->SetPos(Progress);
	}

	delete pBuffer;

	return true;
}

UINT __cdecl SendToEmulatorThread(PVOID pParam)
{
	LINK_THREAD_PARAMS * pltp = (LINK_THREAD_PARAMS *)pParam;
	unsigned Type = 0;

	CProgressCtrl * pProgress = (CProgressCtrl *)CWnd::FromHandle(pltp->hProgressWnd);
	pProgress->SetPos(0);
	pProgress->ShowWindow(SW_SHOW);

	// Is this a single backup file?
	if( pltp->pFiles->GetSize() == 1 )
	{
		CString strFile = pltp->pFiles->GetAt(0);
		strFile.MakeLower();
		
		// Quick & dirty...
		Type = (strFile.Right(1) == 'b') ? 1 : 0;
	}

	try
	{
		TRACE("SendToEmulatorThread - activating link.\n");
		g_pCalculator->m_LinkPort.m_Dir = Inactive;
		g_pCalculator->m_LinkActive = true;

		
		// Send files
		for(int i = 0; i < pltp->pFiles->GetSize(); i++)
			SendFile( Type, pltp->pFiles->GetAt(i), pProgress );

		// Send end of transfer
		if( Type == 0 )
		{
			TIPACKAGEHEADER tiph;
			tiph.Calculator = g_pCalculator->m_Model & 0xF;
			tiph.Command = 0x92;
			WriteBuffer((unsigned char *)&tiph, 4);
			ReadBuffer((unsigned char *)&tiph, 4);
		}

	}
#ifdef _DEBUG
	catch(int nError)
	{
		TRACE("Link error %i trapped\n", nError);
	}
#else
	catch(int)
	{
		// ignore
	}
#endif

	pProgress->ShowWindow(SW_HIDE);
	TRACE("SendToEmulatorThread - deactivating link.\n");
	g_pCalculator->m_LinkSending = g_pCalculator->m_LinkActive = false;

	delete pltp->pFiles;
	delete pParam;

	return 0;
}

//----------------------------------------------------------------------------
//
// Allow the Emulator to send
//
//----------------------------------------------------------------------------
UINT __cdecl ReceiveFromEmulatorThread(PVOID pParam)
{
	//
	// Not yet implemented...
	//
	return 0;
}

void ReadBuffer(unsigned char * pBuffer, unsigned cBuffer)
{
	TRACE("ReadBuffer, %i bytes.\n", cBuffer);

	CLinkPort & lp = g_pCalculator->m_LinkPort;
	lp.m_pBuffer = pBuffer;
	lp.m_cBuffer = cBuffer;
	lp.m_ndxBuffer = 0;
	lp.m_Dir = ExpectSend;

	while( lp.m_Dir != Inactive )
	{
		Sleep(0);

		if( lp.m_Dir == Cancel )
			throw 2;
	}
}

void WriteBuffer(unsigned char * pBuffer, unsigned cBuffer)
{
	TRACE("WriteBuffer, %i bytes.\n", cBuffer);
	
	g_pCalculator->m_LinkSending = true;

	CLinkPort & lp = g_pCalculator->m_LinkPort;
	lp.m_pBuffer = pBuffer;
	lp.m_cBuffer = cBuffer;
	lp.m_ndxBuffer = 0;
	lp.m_Dir = ExpectReceive;

	while( lp.m_Dir != Inactive )
	{
		Sleep(0);

		if( lp.m_Dir == Cancel )
			throw 2;
	}

	g_pCalculator->m_LinkSending = false;
}
	
